import asyncio
import logging
from typing import Any, AsyncGenerator, List, Optional
from uuid import UUID

from core.base import (
    AsyncState,
    EmbeddingProvider,
    KGExtraction,
    PipeType,
    R2RDocumentProcessingError,
    RunLoggingSingleton,
)
from core.base.pipes.base_pipe import AsyncPipe
from core.base.providers import KGProvider

logger = logging.getLogger(__name__)


class KGStoragePipe(AsyncPipe):
    class Input(AsyncPipe.Input):
        message: AsyncGenerator[List[R2RDocumentProcessingError], None]

    def __init__(
        self,
        kg_provider: KGProvider,
        config: AsyncPipe.PipeConfig,
        embedding_provider: Optional[EmbeddingProvider] = None,
        storage_batch_size: int = 1,
        pipe_logger: Optional[RunLoggingSingleton] = None,
        type: PipeType = PipeType.INGESTOR,
        *args,
        **kwargs,
    ):
        """
        Initializes the async knowledge graph storage pipe with necessary components and configurations.
        """
        logger.info(
            f"Initializing an `KGStoragePipe` to store knowledge graph extractions in a graph database."
        )

        super().__init__(
            config,
            type,
            pipe_logger,
            *args,
            **kwargs,
        )
        self.kg_provider = kg_provider
        self.embedding_provider = embedding_provider
        self.storage_batch_size = storage_batch_size

    async def store(
        self,
        kg_extractions: list[KGExtraction],
    ) -> None:
        """
        Stores a batch of knowledge graph extractions in the graph database.
        """
        try:
            return self.kg_provider.upsert_nodes_and_relationships(
                kg_extractions
            )
        except Exception as e:
            error_message = f"Failed to store knowledge graph extractions in the database: {e}"
            logger.error(error_message)
            raise ValueError(error_message)

    async def _run_logic(  # type: ignore
        self,
        input: Input,
        state: AsyncState,
        run_id: UUID,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[List[R2RDocumentProcessingError], None]:
        """
        Executes the async knowledge graph storage pipe: storing knowledge graph extractions in the graph database.
        """

        batch_tasks = []
        kg_batch: list[KGExtraction] = []
        errors = []

        async for kg_extraction in input.message:
            if isinstance(kg_extraction, R2RDocumentProcessingError):
                errors.append(kg_extraction)
                continue

            kg_batch.append(kg_extraction)  # type: ignore
            if len(kg_batch) >= self.storage_batch_size:
                # Schedule the storage task
                batch_tasks.append(
                    asyncio.create_task(
                        self.store(kg_batch.copy()),
                        name=f"kg-store-{self.config.name}",
                    )
                )
                kg_batch.clear()

        if kg_batch:  # Process any remaining extractions
            batch_tasks.append(
                asyncio.create_task(
                    self.store(kg_batch.copy()),
                    name=f"kg-store-{self.config.name}",
                )
            )

        # Wait for all storage tasks to complete
        await asyncio.gather(*batch_tasks)

        for error in errors:
            yield error



import logging
from abc import abstractmethod
from typing import Any, AsyncGenerator, Optional, Union
from uuid import UUID

from core.base import (
    AsyncPipe,
    AsyncState,
    PipeType,
    RunLoggingSingleton,
    VectorSearchResult,
)

logger = logging.getLogger(__name__)


class SearchPipe(AsyncPipe[VectorSearchResult]):
    class SearchConfig(AsyncPipe.PipeConfig):
        name: str = "default_vector_search"
        filters: dict = {}
        search_limit: int = 10

    class Input(AsyncPipe.Input):
        message: Union[AsyncGenerator[str, None], str]

    def __init__(
        self,
        config: AsyncPipe.PipeConfig,
        type: PipeType = PipeType.SEARCH,
        pipe_logger: Optional[RunLoggingSingleton] = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            config,
            type,
            pipe_logger,
            *args,
            **kwargs,
        )

    @abstractmethod
    async def search(
        self,
        query: str,
        search_settings: Any,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[VectorSearchResult, None]:
        pass

    @abstractmethod
    async def _run_logic(
        self,
        input: AsyncPipe.Input,
        state: AsyncState,
        run_id: UUID,
        *args: Any,
        **kwargs,
    ) -> AsyncGenerator[VectorSearchResult, None]:
        pass



from .auth import AuthConfig, AuthProvider
from .base import Provider, ProviderConfig
from .chunking import (
    ChunkingConfig,
    ChunkingProvider,
    R2RChunkingConfig,
    Strategy,
    UnstructuredChunkingConfig,
)
from .crypto import CryptoConfig, CryptoProvider
from .database import (
    DatabaseConfig,
    DatabaseProvider,
    RelationalDBProvider,
    VectorDBProvider,
)
from .embedding import EmbeddingConfig, EmbeddingProvider
from .file import FileConfig, FileProvider
from .kg import KGConfig, KGProvider
from .llm import CompletionConfig, CompletionProvider
from .orchestration import OrchestrationConfig, OrchestrationProvider
from .parsing import OverrideParser, ParsingConfig, ParsingProvider
from .prompt import PromptConfig, PromptProvider

__all__ = [
    # Base provider classes
    "Provider",
    "ProviderConfig",
    # Auth provider
    "AuthConfig",
    "AuthProvider",
    # Chunking provider
    "UnstructuredChunkingConfig",
    "ChunkingConfig",
    "R2RChunkingConfig",
    "ChunkingProvider",
    "Strategy",
    # Crypto provider
    "CryptoConfig",
    "CryptoProvider",
    # Database providers
    "DatabaseConfig",
    "DatabaseProvider",
    "RelationalDBProvider",
    "VectorDBProvider",
    # Embedding provider
    "EmbeddingConfig",
    "EmbeddingProvider",
    # Knowledge Graph provider
    "KGConfig",
    "KGProvider",
    # LLM provider
    "CompletionConfig",
    "CompletionProvider",
    # Orchestration provider
    "OrchestrationConfig",
    "OrchestrationProvider",
    # Parsing provider
    "ParsingConfig",
    "ParsingProvider",
    "OverrideParser",
    # Prompt provider
    "PromptConfig",
    "PromptProvider",
    # File provider
    "FileConfig",
    "FileProvider",
]



from typing import Any, Optional, Sequence, Union

from sqlalchemy import TextClause, text

from .vecs import Client


# TODO: This should be defined at the mixin, not here
def execute_query(
    vx: Client,
    query: Union[str, TextClause],
    params: Optional[dict[str, Any]] = None,
):
    with vx.Session() as sess:
        if isinstance(query, str):
            query = text(query)
        result = sess.execute(query, params or {})
        sess.commit()
        return result


class QueryBuilder:
    def __init__(self, table_name: str):
        self.table_name = table_name
        self.conditions: list[str] = []
        self.params: dict = {}
        self.select_fields = "*"
        self.operation = "SELECT"
        self.limit_value: Optional[int] = None
        self.insert_data: Optional[dict] = None

    def select(self, fields: list[str]):
        self.select_fields = ", ".join(fields)
        return self

    def insert(self, data: dict):
        self.operation = "INSERT"
        self.insert_data = data
        return self

    def delete(self):
        self.operation = "DELETE"
        return self

    def where(self, condition: str, **kwargs):
        self.conditions.append(condition)
        self.params.update(kwargs)
        return self

    def limit(self, value: int):
        self.limit_value = value
        return self

    def build(self):
        if self.operation == "SELECT":
            query = f"SELECT {self.select_fields} FROM {self.table_name}"
        elif self.operation == "INSERT":
            columns = ", ".join(self.insert_data.keys())
            values = ", ".join(f":{key}" for key in self.insert_data.keys())
            query = (
                f"INSERT INTO {self.table_name} ({columns}) VALUES ({values})"
            )
            self.params.update(self.insert_data)
        elif self.operation == "DELETE":
            query = f"DELETE FROM {self.table_name}"
        else:
            raise ValueError(f"Unsupported operation: {self.operation}")

        if self.conditions:
            query += " WHERE " + " AND ".join(self.conditions)

        if self.limit_value is not None and self.operation == "SELECT":
            query += f" LIMIT {self.limit_value}"

        return query, self.params


class DatabaseMixin:
    def _get_table_name(self, base_name: str) -> str:
        raise NotImplementedError("Subclasses must implement this method")

    def execute_query(
        self,
        query: Union[str, TextClause],
        params: Optional[Union[dict[str, Any], Sequence[Any]]] = None,
    ):
        raise NotImplementedError("Subclasses must implement this method")

    def fetch_query(
        self,
        query: Union[str, TextClause],
        params: Optional[Union[dict[str, Any], Sequence[Any]]] = None,
    ):
        raise NotImplementedError("Subclasses must implement this method")

    def fetchrow_query(
        self,
        query: Union[str, TextClause],
        params: Optional[Union[dict[str, Any], Sequence[Any]]] = None,
    ):
        raise NotImplementedError("Subclasses must implement this method")

    def create_table(self):
        raise NotImplementedError("Subclasses must implement this method")



# type: ignore
import json
import logging
import os
from datetime import date, datetime
from decimal import Decimal
from typing import Any, AsyncGenerator, Dict, List, Optional, Tuple
from uuid import UUID

from core.base import KGConfig, KGProvider, R2RException
from core.base.abstractions import (
    Community,
    DocumentFragment,
    Entity,
    KGExtraction,
    RelationshipType,
    Triple,
)

from .graph_queries import (
    GET_CHUNKS_QUERY,
    GET_COMMUNITIES_QUERY,
    GET_ENTITIES_QUERY,
    GET_TRIPLES_BY_SUBJECT_AND_OBJECT_QUERY,
    GET_TRIPLES_QUERY,
    PUT_CHUNKS_QUERY,
    PUT_COMMUNITIES_QUERY,
    PUT_ENTITIES_EMBEDDINGS_QUERY,
    PUT_ENTITIES_QUERY,
    PUT_TRIPLES_QUERY,
    UNIQUE_CONSTRAINTS,
)

logger = logging.getLogger(__name__)


class Neo4jKGProvider(KGProvider):

    def __init__(self, config: KGConfig, *args: Any, **kwargs: Any) -> None:

        try:
            import neo4j
        except ImportError:
            raise ImportError("Please install neo4j: pip install neo4j")

        username = config.user or os.getenv("NEO4J_USER")
        password = config.password or os.getenv("NEO4J_PASSWORD")
        url = config.url or os.getenv("NEO4J_URL")
        database = config.database or os.getenv("NEO4J_DATABASE", "neo4j")

        if not username or not password or not url:
            raise ValueError(
                "Neo4j configuration values are missing. Please set NEO4J_USER, NEO4J_PASSWORD, and NEO4J_URL environment variables."
            )

        self._driver = neo4j.GraphDatabase.driver(
            url, auth=(username, password), **kwargs
        )
        self._async_driver = neo4j.AsyncGraphDatabase.driver(
            url,
            auth=(username, password),
            **kwargs,
        )
        self._database = database
        self.structured_schema = {}
        self.config = config

        self.create_constraints()
        super().__init__(config, *args, **kwargs)

    @property
    def client(self):
        return self._driver

    def create_constraints(self):
        for statement in UNIQUE_CONSTRAINTS:
            self._driver.execute_query(statement)

    def structured_query(self, query: str, param_map: Dict[str, Any] = {}):
        return self._driver.execute_query(query, parameters_=param_map)

    def convert_to_neo4j_compatible(self, value):
        if isinstance(value, (str, int, float, bool)):
            return value
        elif isinstance(value, (datetime, date)):
            return value.isoformat()
        elif isinstance(value, UUID):
            return str(value)
        elif isinstance(value, Decimal):
            return float(value)
        elif isinstance(value, list):
            return value
        elif isinstance(value, dict):
            return json.dumps(value)
        else:
            return str(value)

    def convert_model_list_to_neo4j_compatible(self, model_list):
        return [
            {
                k: self.convert_to_neo4j_compatible(v)
                for k, v in item.dict().items()
            }
            for item in model_list
        ]

    def get_entity_map(
        self, entity_names: list[str] | None = None
    ) -> dict[str, list[Any]]:
        entities = self.get(entity_names)
        triples = self.get_triples(entity_names)
        entity_map = {}
        for entity in entities:
            if entity.name not in entity_map:
                entity_map[entity.name] = {"entities": [], "triples": []}
            entity_map[entity.name]["entities"].append(entity)

        for triple in triples:
            if triple.subject in entity_map:
                entity_map[triple.subject]["triples"].append(triple)
            if triple.object in entity_map:
                entity_map[triple.object]["triples"].append(triple)
        return entity_map

    def batched_import(self, statement, df, batch_size=1000):
        """
        Import a dataframe into Neo4j using a batched approach.
        Parameters: statement is the Cypher query to execute, df is the dataframe to import, and batch_size is the number of rows to import in each batch.
        """
        total = len(df)
        results = []
        for start in range(0, total, batch_size):
            batch = df[start : min(start + batch_size, total)]
            batch = self.convert_model_list_to_neo4j_compatible(batch)
            result = self._driver.execute_query(
                "UNWIND $rows AS value " + statement,
                rows=batch,
                database_=self._database,
            )
            results.append(result)
        return results

    def get_chunks(
        self, chunk_ids: List[str] = None
    ) -> List[DocumentFragment]:
        """
        Get chunks from the graph.
        """
        return self.structured_query(GET_CHUNKS_QUERY, chunk_ids)

    def upsert_chunks(self, chunks: List[DocumentFragment]):
        """
        Upsert chunks into the graph.
        """
        return self.batched_import(PUT_CHUNKS_QUERY, chunks)

    def upsert_entities(
        self, entities: List[Entity], with_embeddings: bool = False
    ):
        """
        Upsert entities into the graph.
        """
        if with_embeddings:
            return self.batched_import(PUT_ENTITIES_EMBEDDINGS_QUERY, entities)
        else:
            return self.batched_import(PUT_ENTITIES_QUERY, entities)

    def upsert_triples(self, triples: List[Triple]):
        """
        Upsert relations into the graph.
        """
        return self.batched_import(PUT_TRIPLES_QUERY, triples)

    def upsert_communities(self, communities: List[Community]):
        """
        Upsert communities into the graph.
        """
        return self.batched_import(PUT_COMMUNITIES_QUERY, communities)

    def get_entities(self, entity_ids: List[str] = []) -> List[Entity]:
        """
        Get entities from the graph.
        """
        neo4j_records = self.structured_query(
            GET_ENTITIES_QUERY, {"entity_ids": entity_ids}
        )
        entities = [
            Entity(
                category=", ".join(list(record["e"]._labels)[1:]),
                **record["e"]._properties,
            )
            for record in neo4j_records.records
        ]
        return entities

    def upsert_nodes_and_relationships(
        self, kg_extractions: list[KGExtraction]
    ) -> Tuple[int, int]:

        all_entities = []
        all_relationships = []
        for extraction in kg_extractions:
            all_entities.extend(list(extraction.entities.values()))
            all_relationships.extend(extraction.triples)

        nodes_upserted = self.upsert_entities(all_entities)
        relationships_upserted = self.upsert_triples(all_relationships)

        return (len(nodes_upserted), len(relationships_upserted))

    def get(self, entity_name: str = None) -> Entity:
        """
        Get entities from the graph.
        """
        if entity_name is None:
            return self.get_entities()
        else:
            return self.get_entities(entity_ids=[entity_name])

    def get_triples(self, triple_ids: list[str] | None = None) -> list[Triple]:
        """
        Get triples from the graph.
        """

        if triple_ids is None:
            neo4j_records = self.structured_query(GET_TRIPLES_QUERY)
        else:
            triple_ids = [triple_id.split("->") for triple_id in triple_ids]
            triple_ids = [
                {
                    "subject": triple_id[0],
                    "predicate": triple_id[1],
                    "object": triple_id[2],
                }
                for triple_id in triple_ids
            ]
            neo4j_records = self.structured_query(
                GET_TRIPLES_BY_SUBJECT_AND_OBJECT_QUERY,
                {"triples": triple_ids},
            )

        triples = [
            Triple(
                subject=record["e1"]._properties["name"],
                predicate=record["rel"].type,
                object=record["e2"]._properties["name"],
                **record["rel"]._properties,
            )
            for record in neo4j_records.records
        ]
        return triples

    def get_community_entities_and_triples(
        self, level: int, community_id: int, include_embeddings: bool = False
    ) -> Tuple[List[Entity], List[Triple]]:
        """
        Get the entities and triples that belong to a community.

        Input:
        - level: The level of the hierarchy.
        - community_id: The ID of the community to get the entities and triples for.
        - include_embeddings: Whether to include the embeddings in the output.

        Output:
        - A tuple of entities and triples that belong to the community.

        """

        # get the entities and triples from the graph
        query = """MATCH (a:__Entity__) - [r] -> (b:__Entity__)
                WHERE a.communityIds[$level] = $community_id
                OR b.communityIds[$level] = $community_id
                RETURN ID(a) AS source_id, a.name AS source, id(b) AS target_id, b.name AS target, a.description AS source_description,
                b.description AS target_description, labels(a) AS source_labels, labels(b) AS target_labels,
                r.description AS relationship_description, r.name AS relationship_name, r.weight AS relationship_weight, ID(r) AS relationship_id
        """

        neo4j_records = self.structured_query(
            query,
            {
                "community_id": int(community_id),
                "level": int(level),
            },
        )

        entities = [
            Entity(
                id=record["source_id"],
                name=record["source"],
                description=record["source_description"],
                category=", ".join(record["source_labels"]),
            )
            for record in neo4j_records.records
        ]

        triples = [
            Triple(
                id=record["relationship_id"],
                subject=record["source"],
                predicate=record["relationship_name"],
                object=record["target"],
                description=record["relationship_description"],
                weight=record["relationship_weight"],
            )
            for record in neo4j_records.records
        ]

        logger.info(
            f"{len(entities)} entities and {len(triples)} triples were retrieved for community {community_id} at level {level}"
        )

        return entities, triples

    def update_extraction_prompt(
        self,
        prompt_provider: Any,
        entity_types: list[Any],
        relations: list[RelationshipType],
    ) -> None:
        pass

    def update_kg_search_prompt(
        self,
        prompt_provider: Any,
        entity_types: list[Any],
        relations: list[RelationshipType],
    ) -> None:
        pass

    def get_communities(self, level: str = None) -> List[Community]:
        """
        Get communities from the graph.
        """
        neo4j_records = self.structured_query(
            GET_COMMUNITIES_QUERY, {"level": level}
        )

        communities = [
            Community(**record["c"]._properties, id=record["c"]["community"])
            for record in neo4j_records.records
        ]
        return communities

    def delete_all_nodes(self):
        self._driver.execute_query("MATCH (a)-[r]->(b) DELETE a, r, b")
        self._driver.execute_query("MATCH (a) DELETE a")

    def delete(
        self,
        entity_names: Optional[List[str]] = None,
        relation_names: Optional[List[str]] = None,
        properties: Optional[dict] = None,
        ids: Optional[List[str]] = None,
    ) -> None:
        pass

    def get_rel_map(
        self,
        graph_nodes: Any,
        depth: int = 2,
        limit: int = 30,
        ignore_rels: Optional[List[str]] = None,
    ) -> List[Triple]:
        pass

    def create_vector_index(
        self, node_type: str, node_property: str, dimension: int
    ) -> None:

        query = f"""
        CREATE VECTOR INDEX `{node_type}_{node_property}` IF NOT EXISTS

        FOR (n:{node_type}) ON n.{node_property}
        OPTIONS {{indexConfig: {{`vector.similarity_function`: 'cosine', `vector.dimensions`:{dimension}}}}}"""

        self.structured_query(query)

    def get_schema(self, refresh: bool = False) -> str:
        return super().get_schema(refresh)

    def retrieve_cache(self, cache_type: str, cache_id: str) -> bool:
        return False

    async def vector_query(
        self, query, **kwargs: Any
    ) -> AsyncGenerator[dict[str, Any], None]:

        query_embedding = kwargs.get("query_embedding", None)
        search_type = kwargs.get("search_type", "__Entity__")
        embedding_type = kwargs.get("embedding_type", "description_embedding")
        property_names = kwargs.get("property_names", ["name", "description"])
        limit = kwargs.get("limit", 10)

        property_names_arr = [
            f"e.{property_name} as {property_name}"
            for property_name in property_names
        ]
        property_names_str = ", ".join(property_names_arr)

        if search_type == "__Relationship__":
            query = f"""
                MATCH () - [e] -> ()
                WHERE e.{embedding_type} IS NOT NULL AND size(e.{embedding_type}) = $dimension
                WITH e, vector.similarity.cosine(e.{embedding_type}, $embedding) AS score
                ORDER BY score DESC LIMIT toInteger($limit)
                RETURN {property_names_str}, score
            """

            query_params = {
                "embedding": query_embedding,
                "dimension": len(query_embedding),
                "limit": limit,
            }

        else:
            query = f"""
                MATCH (e:{search_type})
                WHERE e.{embedding_type} IS NOT NULL AND size(e.{embedding_type}) = $dimension
                WITH e, vector.similarity.cosine(e.{embedding_type}, $embedding) AS score
                ORDER BY score DESC LIMIT toInteger($limit)
                RETURN {property_names_str}, score
            """
            query_params = {
                "embedding": query_embedding,
                "dimension": len(query_embedding),
                "limit": limit,
                "search_type": search_type,
            }

        neo4j_results = self.structured_query(query, query_params)

        # get the descriptions from the neo4j results
        # descriptions = [record['e']._properties[property_name] for record in neo4j_results.records for property_name in property_names]
        # return descriptions, scores
        if search_type == "__Entity__" and len(neo4j_results.records) == 0:
            raise R2RException(
                "No search results found. Please make sure you have run the KG enrichment step before running the search: r2r create-graph and r2r enrich-graph",
                400,
            )

        logger.info(
            f"Neo4j results: Returning {len(neo4j_results.records)} records for query of type {search_type}"
        )

        for record in neo4j_results.records:
            yield {
                property_name: record[property_name]
                for property_name in property_names
            }

    def perform_graph_clustering(
        self, leiden_params: dict
    ) -> Tuple[int, int, set[tuple[int, Any]]]:
        """
        Perform graph clustering on the graph.

        Input:
        - leiden_params: a dictionary that contains the parameters for the graph clustering.

        Output:
        - Total number of communities
        - Total number of hierarchies
        """
        # step 1: drop the graph, if it exists and project the graph again.
        # in this step the vertices that have no edges are not included in the projection.

        GRAPH_EXISTS_QUERY = """
            CALL gds.graph.exists('kg_graph') YIELD exists
            WITH exists
            RETURN CASE WHEN exists THEN true ELSE false END as graphExists;

        """

        result = self.structured_query(GRAPH_EXISTS_QUERY)
        graph_exists = result.records[0]["graphExists"]

        GRAPH_PROJECTION_QUERY = """
            MATCH (s:__Entity__)-[r]->(t:__Entity__)
            RETURN gds.graph.project(
                'kg_graph',
                s,
                t,
        """

        if graph_exists:

            logger.info(f"Graph exists, dropping it")
            GRAPH_DROP_QUERY = (
                "CALL gds.graph.drop('kg_graph') YIELD graphName;"
            )
            result = self.structured_query(GRAPH_DROP_QUERY)

            GRAPH_PROJECTION_QUERY += """
                {
                    sourceNodeProperties: s { },
                    targetNodeProperties: t { },
                    relationshipProperties: r { .weight }
                },
                {
                    relationshipWeightProperty: 'weight',
                    undirectedRelationshipTypes: ['*']
                }
            )
            """
        else:
            GRAPH_PROJECTION_QUERY += """
                {
                    sourceNodeProperties: s {},
                    targetNodeProperties: t {},
                    relationshipProperties: r { .weight }
                },
                {
                    relationshipWeightProperty: 'weight',
                    undirectedRelationshipTypes: ['*']
                }
            )"""

        result = self.structured_query(GRAPH_PROJECTION_QUERY)

        # step 2: run the hierarchical leiden algorithm on the graph.
        # seed_property = leiden_params.get("seed_property", "communityIds")
        write_property = leiden_params.get("write_property", "communityIds")
        random_seed = leiden_params.get("random_seed", 42)
        include_intermediate_communities = leiden_params.get(
            "include_intermediate_communities", True
        )
        max_levels = leiden_params.get("max_levels", 10)
        gamma = leiden_params.get("gamma", 1.0)
        theta = leiden_params.get("theta", 0.01)
        tolerance = leiden_params.get("tolerance", 0.0001)
        min_community_size = leiden_params.get("min_community_size", 1)
        # don't use the seed property for now
        seed_property_config = (
            ""  # f"seedProperty: '{seed_property}'" if graph_exists else ""
        )

        GRAPH_CLUSTERING_QUERY = f"""
            CALL gds.leiden.write('kg_graph', {{
                {seed_property_config}
                writeProperty: '{write_property}',
                randomSeed: {random_seed},
                includeIntermediateCommunities: {include_intermediate_communities},
                maxLevels: {max_levels},
                gamma: {gamma},
                theta: {theta},
                tolerance: {tolerance},
                minCommunitySize: {min_community_size}
            }})
            YIELD communityCount, modularities;
        """

        result = self.structured_query(GRAPH_CLUSTERING_QUERY).records[0]

        community_count: int = result["communityCount"]
        modularities = result["modularities"]

        logger.info(
            f"Performed graph clustering with {community_count} communities and modularities {modularities}"
        )

        COMMUNITY_QUERY = f"""
            MATCH (n)
            WHERE n.communityIds IS NOT NULL
            RETURN DISTINCT
            CASE
                WHEN n.communityIds IS NOT NULL
                THEN toIntegerList(n.communityIds)
                ELSE []
            END AS communityIds
        """

        result = self.structured_query(COMMUNITY_QUERY)

        intermediate_communities = [
            record["communityIds"] for record in result.records
        ]

        intermediate_communities_set = set()
        for community_list in intermediate_communities:
            for level, community_id in enumerate(community_list):
                intermediate_communities_set.add((level, community_id))
        intermediate_communities_set = list(intermediate_communities_set)

        logger.info(
            f"Intermediate communities: {intermediate_communities_set}"
        )

        return (
            community_count,
            len(modularities),
            intermediate_communities_set,
        )



"""Abstractions for documents and their extractions."""

import json
import logging
from datetime import datetime
from enum import Enum
from typing import Optional, Union
from uuid import UUID, uuid4

from pydantic import Field

from .base import R2RSerializable

logger = logging.getLogger(__name__)

DataType = Union[str, bytes]


class DocumentType(str, Enum):
    """Types of documents that can be stored."""

    # Audio
    MP3 = "mp3"

    # CSV
    CSV = "csv"

    # Email
    EML = "eml"
    MSG = "msg"
    P7S = "p7s"

    # EPUB
    EPUB = "epub"

    # Excel
    XLS = "xls"
    XLSX = "xlsx"

    # HTML
    HTML = "html"
    HTM = "htm"

    # Image
    BMP = "bmp"
    HEIC = "heic"
    JPEG = "jpeg"
    PNG = "png"
    TIFF = "tiff"
    JPG = "jpg"
    SVG = "svg"

    # Markdown
    MD = "md"

    # Org Mode
    ORG = "org"

    # Open Office
    ODT = "odt"

    # PDF
    PDF = "pdf"

    # Plain text
    TXT = "txt"
    JSON = "json"

    # PowerPoint
    PPT = "ppt"
    PPTX = "pptx"

    # reStructured Text
    RST = "rst"

    # Rich Text
    RTF = "rtf"

    # TSV
    TSV = "tsv"

    # Video/GIF
    MP4 = "mp4"
    GIF = "gif"

    # Word
    DOC = "doc"
    DOCX = "docx"

    # XML
    XML = "xml"


class Document(R2RSerializable):
    id: UUID = Field(default_factory=uuid4)
    collection_ids: list[UUID]
    user_id: UUID
    type: DocumentType
    metadata: dict

    class Config:
        arbitrary_types_allowed = True
        ignore_extra = False
        json_encoders = {
            UUID: str,
        }


class IngestionStatus(str, Enum):
    """Status of document processing."""

    PENDING = "pending"
    PARSING = "parsing"
    EXTRACTING = "extracting"
    CHUNKING = "chunking"
    EMBEDDING = "embedding"
    STORING = "storing"

    FAILURE = "failure"
    SUCCESS = "success"


class RestructureStatus(str, Enum):
    """Status of document processing."""

    PENDING = "pending"
    PROCESSING = "processing"
    ENRICHING = "enriching"
    FAILURE = "failure"
    SUCCESS = "success"
    ENRICHED = "enriched"
    ENRICHMENT_FAILURE = "enrichment_failure"


class DocumentInfo(R2RSerializable):
    """Base class for document information handling."""

    id: UUID
    collection_ids: list[UUID]
    user_id: UUID
    type: DocumentType
    metadata: dict
    title: Optional[str] = None
    version: str
    size_in_bytes: int
    ingestion_status: IngestionStatus = IngestionStatus.PENDING
    restructuring_status: RestructureStatus = RestructureStatus.PENDING
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    ingestion_attempt_number: Optional[int] = None

    def convert_to_db_entry(self):
        """Prepare the document info for database entry, extracting certain fields from metadata."""
        now = datetime.now()

        return {
            "document_id": self.id,
            "collection_ids": self.collection_ids,
            "user_id": self.user_id,
            "type": self.type,
            "metadata": json.dumps(self.metadata),
            "title": self.title or "N/A",
            "version": self.version,
            "size_in_bytes": self.size_in_bytes,
            "ingestion_status": self.ingestion_status.value,
            "restructuring_status": self.restructuring_status.value,
            "created_at": self.created_at or now,
            "updated_at": self.updated_at or now,
            "ingestion_attempt_number": self.ingestion_attempt_number or 0,
        }


class DocumentExtraction(R2RSerializable):
    """An extraction from a document."""

    id: UUID
    document_id: UUID
    collection_ids: list[UUID]
    user_id: UUID
    data: DataType
    metadata: dict


class DocumentFragment(R2RSerializable):
    """A fragment extracted from a document."""

    id: UUID
    extraction_id: UUID
    document_id: UUID
    user_id: UUID
    collection_ids: list[UUID]
    data: DataType
    metadata: dict



import logging
from datetime import datetime
from typing import Any, AsyncGenerator, Generator, Optional
from uuid import UUID

from core.base import (
    AsyncState,
    CompletionProvider,
    CompletionRecord,
    LLMChatCompletionChunk,
    PipeType,
    PromptProvider,
    format_search_results_for_llm,
    format_search_results_for_stream,
)
from core.base.abstractions import GenerationConfig

from ..abstractions.generator_pipe import GeneratorPipe

logger = logging.getLogger(__name__)


class StreamingSearchRAGPipe(GeneratorPipe):
    VECTOR_SEARCH_STREAM_MARKER = (
        "search"  # TODO - change this to vector_search in next major release
    )
    KG_LOCAL_SEARCH_STREAM_MARKER = "kg_local_search"
    KG_GLOBAL_SEARCH_STREAM_MARKER = "kg_global_search"
    COMPLETION_STREAM_MARKER = "completion"

    def __init__(
        self,
        llm_provider: CompletionProvider,
        prompt_provider: PromptProvider,
        config: GeneratorPipe.PipeConfig,
        type: PipeType = PipeType.GENERATOR,
        *args,
        **kwargs,
    ):
        super().__init__(
            llm_provider,
            prompt_provider,
            config,
            type,
            *args,
            **kwargs,
        )
        self._config: GeneratorPipe.PipeConfig

    @property
    def config(self) -> GeneratorPipe.PipeConfig:
        return self._config

    async def _run_logic(  # type: ignore
        self,
        input: GeneratorPipe.Input,
        state: AsyncState,
        run_id: UUID,
        rag_generation_config: GenerationConfig,
        completion_record: Optional[CompletionRecord] = None,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[str, None]:
        context = ""
        async for query, search_results in input.message:
            result = format_search_results_for_stream(search_results)
            yield result
            gen_context = format_search_results_for_llm(search_results)
            context += gen_context

        messages = self.prompt_provider._get_message_payload(
            system_prompt_name=self.config.system_prompt,
            task_prompt_name=self.config.task_prompt,
            task_inputs={"query": query, "context": context},
        )
        yield f"<{self.COMPLETION_STREAM_MARKER}>"

        response = ""
        for chunk in self.llm_provider.get_completion_stream(
            messages=messages, generation_config=rag_generation_config
        ):
            chunk_txt = StreamingSearchRAGPipe._process_chunk(chunk)
            response += chunk_txt
            yield chunk_txt

        yield f"</{self.COMPLETION_STREAM_MARKER}>"
        if not completion_record:
            raise ValueError(
                "Completion record is expected in the streaming RAG pipe and is used for logging."
            )
        completion_record.search_results = search_results
        completion_record.llm_response = response
        completion_record.completion_end_time = datetime.now()
        await self.log_completion_record(run_id, completion_record)

    async def _yield_chunks(
        self,
        start_marker: str,
        chunks: Generator[str, None, None],
        end_marker: str,
    ) -> AsyncGenerator[str, None]:
        yield start_marker
        for chunk in chunks:
            yield chunk
        yield end_marker

    @staticmethod
    def _process_chunk(chunk: LLMChatCompletionChunk) -> str:
        return chunk.choices[0].delta.content or ""

    async def log_completion_record(
        self, run_id: UUID, completion_record: CompletionRecord
    ):
        await self.enqueue_log(
            run_id=run_id,
            key="completion_record",
            value=completion_record.to_json(),
        )



import uuid
from typing import Any, Dict, Optional


class BaseTelemetryEvent:
    def __init__(self, event_type: str, properties: Dict[str, Any]):
        self.event_type = event_type
        self.properties = properties
        self.event_id = str(uuid.uuid4())


class DailyActiveUserEvent(BaseTelemetryEvent):
    def __init__(self, user_id: str):
        super().__init__("DailyActiveUser", {"user_id": user_id})


class FeatureUsageEvent(BaseTelemetryEvent):
    def __init__(
        self,
        user_id: str,
        feature: str,
        properties: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            "FeatureUsage",
            {
                "user_id": user_id,
                "feature": feature,
                "properties": properties or {},
            },
        )


class ErrorEvent(BaseTelemetryEvent):
    def __init__(
        self,
        user_id: str,
        endpoint: str,
        error_message: str,
        properties: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            "Error",
            {
                "user_id": user_id,
                "endpoint": endpoint,
                "error_message": error_message,
                "properties": properties or {},
            },
        )


class RequestLatencyEvent(BaseTelemetryEvent):
    def __init__(
        self,
        endpoint: str,
        latency: float,
        properties: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            "RequestLatency",
            {
                "endpoint": endpoint,
                "latency": latency,
                "properties": properties or {},
            },
        )


class GeographicDistributionEvent(BaseTelemetryEvent):
    def __init__(
        self,
        user_id: str,
        country: str,
        properties: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            "GeographicDistribution",
            {
                "user_id": user_id,
                "country": country,
                "properties": properties or {},
            },
        )


class SessionDurationEvent(BaseTelemetryEvent):
    def __init__(
        self,
        user_id: str,
        duration: float,
        properties: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            "SessionDuration",
            {
                "user_id": user_id,
                "duration": duration,
                "properties": properties or {},
            },
        )


class UserPathEvent(BaseTelemetryEvent):
    def __init__(
        self,
        user_id: str,
        path: str,
        properties: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            "UserPath",
            {"user_id": user_id, "path": path, "properties": properties or {}},
        )



import asyncclick as click
from asyncclick import pass_context

from cli.command_group import cli
from cli.utils.param_types import JSON
from cli.utils.timer import timer


@cli.command()
@click.option(
    "--query", prompt="Enter your search query", help="The search query"
)
# VectorSearchSettings
@click.option(
    "--use-vector-search",
    is_flag=True,
    default=True,
    help="Whether to use vector search",
)
@click.option(
    "--filters",
    type=JSON,
    help="Filters to apply to the vector search as a JSON",
)
@click.option(
    "--search-limit", default=None, help="Number of search results to return"
)
@click.option(
    "--use-hybrid-search", is_flag=True, help="Perform hybrid search"
)
@click.option(
    "--selected-collection-ids",
    type=JSON,
    help="Collection IDs to search for as a JSON",
)
# KGSearchSettings
@click.option(
    "--use-kg-search", is_flag=True, help="Use knowledge graph search"
)
@click.option("--kg-search-type", default=None, help="Local or Global")
@click.option("--kg-search-level", default=None, help="Level of KG search")
@click.option(
    "--kg-search-generation-config",
    type=JSON,
    help="KG search generation config",
)
@click.option(
    "--entity-types", type=JSON, help="Entity types to search for as a JSON"
)
@click.option(
    "--relationships", type=JSON, help="Relationships to search for as a JSON"
)
@click.option(
    "--max-community-description-length",
    type=JSON,
    help="Max community description length",
)
@click.option(
    "--max-llm-queries-for-global-search", type=JSON, help="Max community size"
)
@click.option(
    "--search-strategy",
    type=str,
    help="Vanilla search or complex search method like query fusion or HyDE.",
)
@click.option("--local-search-limits", type=JSON, help="Local search limits")
@pass_context
def search(ctx, query, **kwargs):
    """Perform a search query."""
    client = ctx.obj
    vector_search_settings = {
        k: v
        for k, v in kwargs.items()
        if k
        in [
            "use_vector_search",
            "filters",
            "search_limit",
            "use_hybrid_search",
            "selected_collection_ids",
            "search_strategy",
        ]
        and v is not None
    }

    kg_search_settings = {
        k: v
        for k, v in kwargs.items()
        if k
        in [
            "use_kg_search",
            "kg_search_type",
            "kg_search_level",
            "generation_config",
            "entity_types",
            "relationships",
            "max_community_description_length",
            "max_llm_queries_for_global_search",
            "local_search_limits",
        ]
        and v is not None
    }

    with timer():
        results = client.search(
            query,
            vector_search_settings,
            kg_search_settings,
        )

        if isinstance(results, dict) and "results" in results:
            results = results["results"]

        if "vector_search_results" in results:
            click.echo("Vector search results:")
            for result in results["vector_search_results"]:
                click.echo(result)

        if "kg_search_results" in results and results["kg_search_results"]:
            click.echo("KG search results:")
            for result in results["kg_search_results"]:
                click.echo(result)


@cli.command()
@click.option("--query", prompt="Enter your query", help="The query for RAG")
# RAG Generation Config
@click.option("--stream", is_flag=True, help="Stream the RAG response")
@click.option("--rag-model", default=None, help="Model for RAG")
# Vector Search Settings
@click.option(
    "--use-vector-search", is_flag=True, default=True, help="Use vector search"
)
@click.option("--filters", type=JSON, help="Search filters as JSON")
@click.option(
    "--search-limit", default=10, help="Number of search results to return"
)
@click.option(
    "--use-hybrid-search", is_flag=True, help="Perform hybrid search"
)
@click.option(
    "--selected-collection-ids",
    type=JSON,
    help="Collection IDs to search for as a JSON",
)
# KG Search Settings
@click.option(
    "--use-kg-search", is_flag=True, help="Use knowledge graph search"
)
@click.option("--kg-search-type", default="local", help="Local or Global")
@click.option(
    "--kg-search-level",
    default=None,
    help="Level of cluster to use for Global KG search",
)
@click.option("--kg-search-model", default=None, help="Model for KG agent")
@click.option(
    "--entity-types", type=JSON, help="Entity types to search for as a JSON"
)
@click.option(
    "--relationships", type=JSON, help="Relationships to search for as a JSON"
)
@click.option(
    "--max-community-description-length",
    type=int,
    help="Max community description length",
)
@click.option(
    "--max-llm-queries-for-global-search", type=int, help="Max community size"
)
@click.option(
    "--search-strategy",
    type=str,
    default="vanilla",
    help="Vanilla RAG or complex method like query fusion or HyDE.",
)
@click.option("--local-search-limits", type=JSON, help="Local search limits")
@pass_context
def rag(ctx, query, **kwargs):
    """Perform a RAG query."""
    client = ctx.obj
    rag_generation_config = {
        "stream": kwargs.get("stream", False),
    }
    if kwargs.get("rag_model"):
        rag_generation_config["model"] = kwargs["rag_model"]

    vector_search_settings = {
        k: v
        for k, v in kwargs.items()
        if k
        in [
            "use_vector_search",
            "filters",
            "search_limit",
            "use_hybrid_search",
            "selected_collection_ids",
            "search_strategy",
        ]
        and v is not None
    }

    kg_search_settings = {
        k: v
        for k, v in kwargs.items()
        if k
        in [
            "use_kg_search",
            "kg_search_type",
            "kg_search_level",
            "kg_search_model",
            "entity_types",
            "relationships",
            "max_community_description_length",
            "max_llm_queries_for_global_search",
            "local_search_limits",
        ]
        and v is not None
    }

    if kg_search_settings.get("kg_search_model"):
        kg_search_settings["generation_config"] = {
            "model": kg_search_settings.pop("kg_search_model")
        }

    with timer():
        response = client.rag(
            query,
            rag_generation_config,
            vector_search_settings,
            kg_search_settings,
        )

        if rag_generation_config.get("stream"):
            for chunk in response:
                click.echo(chunk, nl=False)
            click.echo()
        else:
            click.echo(response)


# TODO: Implement agent



import os
from typing import AsyncGenerator

from core.base.parsers.base_parser import AsyncParser
from core.parsers.media.openai_helpers import process_audio_with_openai


class AudioParser(AsyncParser[bytes]):
    """A parser for audio data."""

    def __init__(
        self, api_base: str = "https://api.openai.com/v1/audio/transcriptions"
    ):
        self.api_base = api_base
        self.openai_api_key = os.environ.get("OPENAI_API_KEY")

    async def ingest(  # type: ignore
        self, data: bytes, chunk_size: int = 1024
    ) -> AsyncGenerator[str, None]:
        """Ingest audio data and yield a transcription."""
        temp_audio_path = "temp_audio.wav"
        with open(temp_audio_path, "wb") as f:
            f.write(data)
        try:
            transcription_text = process_audio_with_openai(
                open(temp_audio_path, "rb"), self.openai_api_key  # type: ignore
            )

            # split text into small chunks and yield them
            for i in range(0, len(transcription_text), chunk_size):
                text = transcription_text[i : i + chunk_size]
                if text and text != "":
                    yield text
        finally:
            os.remove(temp_audio_path)



from pathlib import Path

import toml

from sdk import *
from shared import *

pyproject_path = Path(__file__).parent.parent / "pyproject.toml"
pyproject_data = toml.load(pyproject_path)
__version__ = pyproject_data["tool"]["poetry"]["version"]


__all__ = [
    # R2R SDK
    "R2RAsyncClient",
    "R2RClient",
    "__version__",
]

try:
    import core
    from core import *

    __all__ += core.__all__
except ImportError:
    # Core dependencies not installed
    pass


# Add a function to get the version
def get_version():
    return __version__



from abc import ABC, abstractmethod
from io import BytesIO
from typing import BinaryIO, Optional
from uuid import UUID

from .base import Provider, ProviderConfig


class FileConfig(ProviderConfig):
    def validate_config(self) -> None:
        if self.provider not in self.supported_providers:
            raise ValueError(f"Provider '{self.provider}' is not supported.")

    @property
    def supported_providers(self) -> list[str]:
        return ["postgres"]


class FileProvider(Provider, ABC):
    @abstractmethod
    async def store_file(
        self,
        document_id: UUID,
        file_name: str,
        file_content: BytesIO,
        file_type: Optional[str] = None,
    ) -> None:
        pass

    @abstractmethod
    async def retrieve_file(
        self, document_id: UUID
    ) -> Optional[tuple[str, BinaryIO, int]]:
        pass

    @abstractmethod
    async def delete_file(self, document_id: UUID) -> bool:
        pass

    @abstractmethod
    async def get_files_overview(
        self,
        filter_document_ids: Optional[list[UUID]] = None,
        filter_file_names: Optional[list[str]] = None,
        offset: int = 0,
        limit: int = 100,
    ) -> list[dict]:
        pass



import json
import os
import platform
import subprocess
import sys
from importlib.metadata import version as get_version

import asyncclick as click
from asyncclick import pass_context
from dotenv import load_dotenv

from cli.command_group import cli
from cli.utils.docker_utils import (
    bring_down_docker_compose,
    remove_r2r_network,
    run_docker_serve,
    run_local_serve,
    wait_for_container_health,
)
from cli.utils.timer import timer


@cli.command()
@pass_context
def health(ctx):
    """Check the health of the server."""
    client = ctx.obj
    with timer():
        response = client.health()

    click.echo(response)


@cli.command()
@pass_context
def server_stats(ctx):
    client = ctx.obj
    """Check the server stats."""
    with timer():
        response = client.server_stats()

    click.echo(response)


@cli.command()
@click.option(
    "--offset", default=None, help="Pagination offset. Default is None."
)
@click.option(
    "--limit", default=None, help="Pagination limit. Defaults to 100."
)
@click.option("--run-type-filter", help="Filter for log types")
@pass_context
def logs(ctx, run_type_filter, offset, limit):
    """Retrieve logs with optional type filter."""
    client = ctx.obj
    with timer():
        response = client.logs(
            offset=offset, limit=limit, run_type_filter=run_type_filter
        )

    for log in response["results"]:
        click.echo(f"Run ID: {log['run_id']}")
        click.echo(f"Run Type: {log['run_type']}")
        click.echo(f"Timestamp: {log['timestamp']}")
        click.echo(f"User ID: {log['user_id']}")
        click.echo("Entries:")
        for entry in log["entries"]:
            click.echo(f"  - {entry['key']}: {entry['value'][:100]}")
        click.echo("---")

    click.echo(f"Total runs: {len(response['results'])}")


@cli.command()
@click.option(
    "--volumes",
    is_flag=True,
    help="Remove named volumes declared in the `volumes` section of the Compose file",
)
@click.option(
    "--remove-orphans",
    is_flag=True,
    help="Remove containers for services not defined in the Compose file",
)
@click.option("--project-name", default="r2r", help="Project name for Docker")
def docker_down(volumes, remove_orphans, project_name):
    """Bring down the Docker Compose setup and attempt to remove the network if necessary."""
    result = bring_down_docker_compose(project_name, volumes, remove_orphans)
    remove_r2r_network()

    if result != 0:
        click.echo(
            "An error occurred while bringing down the Docker Compose setup. Attempting to remove the network..."
        )
    else:
        click.echo("Docker Compose setup has been successfully brought down.")


@cli.command()
def generate_report():
    """Generate a system report including R2R version, Docker info, and OS details."""

    # Get R2R version
    from importlib.metadata import version

    report = {"r2r_version": version("r2r")}

    # Get Docker info
    try:
        subprocess.run(
            ["docker", "version"], check=True, capture_output=True, timeout=5
        )

        docker_ps_output = subprocess.check_output(
            ["docker", "ps", "--format", "{{.ID}}\t{{.Names}}\t{{.Status}}"],
            text=True,
            timeout=5,
        ).strip()
        report["docker_ps"] = [
            dict(zip(["id", "name", "status"], line.split("\t")))
            for line in docker_ps_output.split("\n")
            if line
        ]

        docker_network_output = subprocess.check_output(
            ["docker", "network", "ls", "--format", "{{.ID}}\t{{.Name}}"],
            text=True,
            timeout=5,
        ).strip()
        networks = [
            dict(zip(["id", "name"], line.split("\t")))
            for line in docker_network_output.split("\n")
            if line
        ]

        report["docker_subnets"] = []
        for network in networks:
            inspect_output = subprocess.check_output(
                [
                    "docker",
                    "network",
                    "inspect",
                    network["id"],
                    "--format",
                    "{{range .IPAM.Config}}{{.Subnet}}{{end}}",
                ],
                text=True,
                timeout=5,
            ).strip()
            if subnet := inspect_output:
                network["subnet"] = subnet
                report["docker_subnets"].append(network)

    except subprocess.CalledProcessError as e:
        report["docker_error"] = f"Error running Docker command: {e}"
    except FileNotFoundError:
        report["docker_error"] = (
            "Docker command not found. Is Docker installed and in PATH?"
        )
    except subprocess.TimeoutExpired:
        report["docker_error"] = (
            "Docker command timed out. Docker might be unresponsive."
        )

    # Get OS information
    report["os_info"] = {
        "system": platform.system(),
        "release": platform.release(),
        "version": platform.version(),
        "machine": platform.machine(),
        "processor": platform.processor(),
    }

    click.echo("System Report:")
    click.echo(json.dumps(report, indent=2))


@cli.command()
@click.option("--host", default="0.0.0.0", help="Host to run the server on")
@click.option("--port", default=7272, help="Port to run the server on")
@click.option("--docker", is_flag=True, help="Run using Docker")
@click.option(
    "--exclude-neo4j", default=False, help="Exclude Neo4j from Docker setup"
)
@click.option(
    "--exclude-ollama", default=True, help="Exclude Ollama from Docker setup"
)
@click.option(
    "--exclude-postgres",
    default=False,
    help="Exclude Postgres from Docker setup",
)
@click.option(
    "--exclude-hatchet",
    default=False,
    help="Exclude Hatchet from Docker setup",
)
@click.option("--project-name", default="r2r", help="Project name for Docker")
@click.option("--image", help="Docker image to use")
@click.option(
    "--config-name", default=None, help="Name of the R2R configuration to use"
)
@click.option(
    "--config-path",
    default=None,
    help="Path to a custom R2R configuration file",
)
@click.option(
    "--build",
    is_flag=True,
    default=False,
    help="Run in debug mode. Only for development.",
)
@click.option(
    "--dev",
    is_flag=True,
    default=False,
    help="Run in development mode",
)
@click.option(
    "--image-env",
    default="prod",
    help="Which dev environment to pull the image from?",
)
async def serve(
    host,
    port,
    docker,
    exclude_neo4j,
    exclude_ollama,
    exclude_postgres,
    exclude_hatchet,
    project_name,
    image,
    config_name,
    config_path,
    build,
    dev,
    image_env,
):
    """Start the R2R server."""
    load_dotenv()
    if image and image_env:
        click.echo(
            "WARNING: Both `image` and `image_env` were provided. Using `image`."
        )
    if not image and docker:
        r2r_version = get_version("r2r")

        version_specific_image = f"ragtoriches/{image_env}:{r2r_version}"
        latest_image = f"ragtoriches/{image_env}:latest"

        def image_exists(img):
            try:
                subprocess.run(
                    ["docker", "manifest", "inspect", img],
                    check=True,
                    capture_output=True,
                    text=True,
                )
                return True
            except subprocess.CalledProcessError:
                return False

        if image_exists(version_specific_image):
            click.echo(f"Using image: {version_specific_image}")
            image = version_specific_image
        elif image_exists(latest_image):
            click.echo(
                f"Version-specific image not found. Using latest: {latest_image}"
            )
            image = latest_image
        else:
            click.echo(
                f"Neither {version_specific_image} nor {latest_image} found in remote registry. Confirm the sanity of your output for `docker manifest inspect ragtoriches/{version_specific_image}` and  `docker manifest inspect ragtoriches/{latest_image}`."
            )
            click.echo(
                "Please pull the required image or build it using the --build flag."
            )
            raise click.Abort()

    if docker:
        os.environ["R2R_IMAGE"] = image

    if build:
        subprocess.run(
            [
                "docker",
                "build",
                "-t",
                image,
                "-f",
                f"Dockerfile{'.dev' if dev else ''}",
                ".",
            ],
            check=True,
        )

    if config_path:
        config_path = os.path.abspath(config_path)

        # For Windows, convert backslashes to forward slashes and prepend /host_mnt/
        if platform.system() == "Windows":
            config_path = "/host_mnt/" + config_path.replace(
                "\\", "/"
            ).replace(":", "")

    if docker:
        run_docker_serve(
            host,
            port,
            exclude_neo4j,
            exclude_ollama,
            exclude_postgres,
            exclude_hatchet,
            project_name,
            image,
            config_name,
            config_path,
        )
        if (
            "pytest" in sys.modules
            or "unittest" in sys.modules
            or os.environ.get("PYTEST_CURRENT_TEST")
        ):
            click.echo("Test environment detected. Skipping browser open.")
        else:
            # Open browser after Docker setup is complete
            import webbrowser

            click.echo("Waiting for all services to become healthy...")

            if not wait_for_container_health(project_name, "r2r"):
                click.secho(
                    "r2r container failed to become healthy.", fg="red"
                )
                return

            traefik_port = os.environ.get("R2R_DASHBOARD_PORT", "80")
            url = f"http://localhost:{traefik_port}"

            click.secho(f"Navigating to R2R application at {url}.", fg="blue")
            webbrowser.open(url)
    else:
        await run_local_serve(host, port, config_name, config_path)


@cli.command()
def update():
    """Update the R2R package to the latest version."""
    try:
        cmd = [sys.executable, "-m", "pip", "install", "--upgrade", "r2r"]

        click.echo("Updating R2R...")
        result = subprocess.run(
            cmd, check=True, capture_output=True, text=True
        )
        click.echo(result.stdout)
        click.echo("R2R has been successfully updated.")
    except subprocess.CalledProcessError as e:
        click.echo(f"An error occurred while updating R2R: {e}")
        click.echo(e.stderr)
    except Exception as e:
        click.echo(f"An unexpected error occurred: {e}")


@cli.command()
def version():
    """Print the version of R2R."""
    from importlib.metadata import version

    click.echo(version("r2r"))



"""Base classes for knowledge graph providers."""

import logging
from abc import ABC, abstractmethod
from typing import Any, Optional, Tuple

from ..abstractions import (
    Entity,
    GenerationConfig,
    KGCreationSettings,
    KGEnrichmentSettings,
    KGExtraction,
    KGSearchSettings,
    RelationshipType,
    Triple,
)
from .base import ProviderConfig

logger = logging.getLogger(__name__)


class KGConfig(ProviderConfig):
    """A base KG config class"""

    provider: Optional[str] = None
    user: Optional[str] = None
    password: Optional[str] = None
    url: Optional[str] = None
    database: Optional[str] = None

    batch_size: Optional[int] = 1
    kg_store_path: Optional[str] = None
    kg_enrichment_settings: KGEnrichmentSettings = KGEnrichmentSettings()
    kg_creation_settings: KGCreationSettings = KGCreationSettings()
    kg_search_settings: KGSearchSettings = KGSearchSettings()

    def validate_config(self) -> None:
        if self.provider not in self.supported_providers:
            raise ValueError(f"Provider '{self.provider}' is not supported.")

    @property
    def supported_providers(self) -> list[str]:
        return ["neo4j", "local"]


class KGProvider(ABC):
    """An abstract class to provide a common interface for Knowledge Graphs."""

    def __init__(self, config: KGConfig) -> None:
        if not isinstance(config, KGConfig):
            raise ValueError(
                "KGProvider must be initialized with a `KGConfig`."
            )
        logger.info(f"Initializing KG provider with config: {config}")
        self.config = config
        self.validate_config()

    def validate_config(self) -> None:
        self.config.validate_config()

    @property
    @abstractmethod
    def client(self) -> Any:
        """Get client."""
        pass

    @abstractmethod
    def get(self, subj: str) -> list[list[str]]:
        """Abstract method to get triplets."""
        pass

    @abstractmethod
    def get_rel_map(
        self,
        subjs: Optional[list[str]] = None,
        depth: int = 2,
        limit: int = 30,
    ) -> dict[str, list[list[str]]]:
        """Abstract method to get depth-aware rel map."""
        pass

    @abstractmethod
    def upsert_entities(self, entities: list[Entity], *args, **kwargs) -> None:
        """Abstract method to add triplet."""
        pass

    @abstractmethod
    def upsert_triples(self, triples: list[Triple]) -> None:
        """Abstract method to add triplet."""
        pass

    @abstractmethod
    def get_entities(
        self,
        entity_ids: list[str] | None = None,
        with_description: bool = False,
    ) -> list[Entity]:
        """Abstract method to get entities."""
        pass

    @abstractmethod
    def get_triples(self, triple_ids: list[str] | None = None) -> list[Triple]:
        """Abstract method to get triples."""
        pass

    @abstractmethod
    def upsert_nodes_and_relationships(
        self, kg_extractions: list[KGExtraction]
    ) -> None:
        """Abstract method to add triplet."""
        pass

    @abstractmethod
    def delete(self, subj: str, rel: str, obj: str) -> None:
        """Abstract method to delete triplet."""
        pass

    @abstractmethod
    def get_schema(self, refresh: bool = False) -> str:
        """Abstract method to get the schema of the graph store."""
        pass

    @abstractmethod
    def structured_query(
        self, query: str, param_map: Optional[dict[str, Any]] = None
    ) -> Any:
        """Abstract method to query the graph store with statement and parameters."""
        if param_map is None:
            param_map = {}

    @abstractmethod
    async def vector_query(
        self, query, **kwargs: Any
    ) -> Tuple[list[Entity], list[float]]:
        """Abstract method to query the graph store with a vector store query."""

    # TODO - Type this method.
    @abstractmethod
    def update_extraction_prompt(
        self,
        prompt_provider: Any,
        entity_types: list[Any],
        relationship_types: list[RelationshipType],
    ):
        """Abstract method to update the KG extraction prompt."""
        pass

    # TODO - Type this method.
    @abstractmethod
    def update_kg_search_prompt(
        self,
        prompt_provider: Any,
        entity_types: list[Any],
        relationship_types: list[RelationshipType],
    ):
        """Abstract method to update the KG agent prompt."""
        pass

    @abstractmethod
    def create_vector_index(
        self, node_type: str, node_property: str, dimension: int
    ) -> None:
        """Abstract method to create a vector index."""
        pass


def escape_braces(s: str) -> str:
    """
    Escape braces in a string.
    This is a placeholder function - implement the actual logic as needed.
    """
    # Implement your escape_braces logic here
    return s.replace("{", "{{").replace("}", "}}")



from typing import Any, Optional

from pydantic import BaseModel, Field

from shared.abstractions import KGSearchResult, Message, VectorSearchResult
from shared.api.models.base import ResultsWrapper


class SearchResponse(BaseModel):
    vector_search_results: list[VectorSearchResult] = Field(
        ...,
        description="List of vector search results",
    )
    kg_search_results: Optional[list[KGSearchResult]] = Field(
        None,
        description="Knowledge graph search results, if applicable",
    )

    class Config:
        json_schema_extra = {
            "example": {
                "vector_search_results": [
                    VectorSearchResult.Config.json_schema_extra,
                ],
                "kg_search_results": [
                    KGSearchResult.Config.json_schema_extra,
                ],
            }
        }


class RAGResponse(BaseModel):
    completion: Any = Field(
        ...,
        description="The generated completion from the RAG process",
    )
    search_results: SearchResponse = Field(
        ...,
        description="The search results used for the RAG process",
    )

    class Config:
        json_schema_extra = {
            "example": {
                "completion": {
                    "id": "chatcmpl-example123",
                    "choices": [
                        {
                            "finish_reason": "stop",
                            "index": 0,
                            "logprobs": None,
                            "message": {
                                "content": "Paris is the capital of France.",
                                "role": "assistant",
                            },
                        }
                    ],
                },
                "search_results": {
                    "vector_search_results": [
                        VectorSearchResult.Config.json_schema_extra,
                    ],
                    "kg_search_results": [
                        KGSearchResult.Config.json_schema_extra,
                    ],
                },
            }
        }


class RAGAgentResponse(BaseModel):
    messages: list[Message] = Field(
        ..., description="List of messages in the RAG agent response"
    )

    class Config:
        json_schema_extra = {
            "example": [
                {
                    "role": "system",
                    "content": "## You are a helpful assistant that can search for information.\n\nWhen asked a question, perform a search to find relevant information and provide a response.\n\nThe response should contain line-item attributions to relevent search results, and be as informative if possible.\nIf no relevant results are found, then state that no results were found.\nIf no obvious question is present, then do not carry out a search, and instead ask for clarification.",
                    "name": None,
                    "function_call": None,
                    "tool_calls": None,
                },
                {
                    "role": "system",
                    "content": "You are a helpful assistant.",
                    "name": None,
                    "function_call": None,
                    "tool_calls": None,
                },
                {
                    "role": "user",
                    "content": "Who is the greatest philospher of all time?",
                    "name": None,
                    "function_call": None,
                    "tool_calls": None,
                },
                {
                    "role": "assistant",
                    "content": "Aristotle is widely considered the greatest philospher of all time.",
                    "name": None,
                    "function_call": None,
                    "tool_calls": None,
                },
                {
                    "role": "user",
                    "content": "Can you tell me more about him?",
                    "name": None,
                    "function_call": None,
                    "tool_calls": None,
                },
                {
                    "role": "assistant",
                    "content": None,
                    "name": None,
                    "function_call": {
                        "name": "search",
                        "arguments": '{"query":"Aristotle biography"}',
                    },
                    "tool_calls": None,
                },
                {
                    "role": "function",
                    "content": "1. Aristotle[A] (Greek: Ἀριστοτέλης Aristotélēs, pronounced [aristotélɛːs]; 384–322 BC) was an Ancient Greek philosopher and polymath. His writings cover a broad range of subjects spanning the natural sciences, philosophy, linguistics, economics, politics, psychology, and the arts. As the founder of the Peripatetic school of philosophy in the Lyceum in Athens, he began the wider Aristotelian tradition that followed, which set the groundwork for the development of modern science.\n2. Aristotle[A] (Greek: Ἀριστοτέλης Aristotélēs, pronounced [aristotélɛːs]; 384–322 BC) was an Ancient Greek philosopher and polymath. His writings cover a broad range of subjects spanning the natural sciences, philosophy, linguistics, economics, politics, psychology, and the arts. As the founder of the Peripatetic school of philosophy in the Lyceum in Athens, he began the wider Aristotelian tradition that followed, which set the groundwork for the development of modern science.\n3. Aristotle was born in 384 BC[C] in Stagira, Chalcidice,[2] about 55 km (34 miles) east of modern-day Thessaloniki.[3][4] His father, Nicomachus, was the personal physician to King Amyntas of Macedon. While he was young, Aristotle learned about biology and medical information, which was taught by his father.[5] Both of Aristotle's parents died when he was about thirteen, and Proxenus of Atarneus became his guardian.[6] Although little information about Aristotle's childhood has survived, he probably spent\n4. Aristotle was born in 384 BC[C] in Stagira, Chalcidice,[2] about 55 km (34 miles) east of modern-day Thessaloniki.[3][4] His father, Nicomachus, was the personal physician to King Amyntas of Macedon. While he was young, Aristotle learned about biology and medical information, which was taught by his father.[5] Both of Aristotle's parents died when he was about thirteen, and Proxenus of Atarneus became his guardian.[6] Although little information about Aristotle's childhood has survived, he probably spent\n5. Life\nIn general, the details of Aristotle's life are not well-established. The biographies written in ancient times are often speculative and historians only agree on a few salient points.[B]\n",
                    "name": "search",
                    "function_call": None,
                    "tool_calls": None,
                },
                {
                    "role": "assistant",
                    "content": "Aristotle (384–322 BC) was an Ancient Greek philosopher and polymath whose contributions have had a profound impact on various fields of knowledge. Here are some key points about his life and work:\n\n1. **Early Life**: Aristotle was born in 384 BC in Stagira, Chalcidice, which is near modern-day Thessaloniki, Greece. His father, Nicomachus, was the personal physician to King Amyntas of Macedon, which exposed Aristotle to medical and biological knowledge from a young age [C].\n\n2. **Education and Career**: After the death of his parents, Aristotle was sent to Athens to study at Plato's Academy, where he remained for about 20 years. After Plato's death, Aristotle left Athens and eventually became the tutor of Alexander the Great [C].\n\n3. **Philosophical Contributions**: Aristotle founded the Lyceum in Athens, where he established the Peripatetic school of philosophy. His works cover a wide range of subjects, including metaphysics, ethics, politics, logic, biology, and aesthetics. His writings laid the groundwork for many modern scientific and philosophical inquiries [A].\n\n4. **Legacy**: Aristotle's influence extends beyond philosophy to the natural sciences, linguistics, economics, and psychology. His method of systematic observation and analysis has been foundational to the development of modern science [A].\n\nAristotle's comprehensive approach to knowledge and his systematic methodology have earned him a lasting legacy as one of the greatest philosophers of all time.\n\nSources:\n- [A] Aristotle's broad range of writings and influence on modern science.\n- [C] Details about Aristotle's early life and education.",
                    "name": None,
                    "function_call": None,
                    "tool_calls": None,
                },
            ]
        }


# Create wrapped versions of the responses
WrappedSearchResponse = ResultsWrapper[SearchResponse]
WrappedRAGResponse = ResultsWrapper[RAGResponse]
WrappedRAGAgentResponse = ResultsWrapper[RAGAgentResponse]



import uuid
from typing import Optional

from fastapi import Body, Depends, Path
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm
from pydantic import EmailStr

from core.base.api.models import (
    GenericMessageResponse,
    WrappedGenericMessageResponse,
    WrappedTokenResponse,
    WrappedUserResponse,
)
from core.base.providers import OrchestrationProvider

from ..services.auth_service import AuthService
from .base_router import BaseRouter, RunType

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")


class AuthRouter(BaseRouter):
    def __init__(
        self,
        auth_service: AuthService,
        run_type: RunType = RunType.INGESTION,
        orchestration_provider: Optional[OrchestrationProvider] = None,
    ):
        super().__init__(auth_service, run_type, orchestration_provider)
        self.service: AuthService = auth_service  # for type hinting

    def _register_workflows(self):
        pass

    def _load_openapi_extras(self):
        return {}

    def _setup_routes(self):
        @self.router.post("/register", response_model=WrappedUserResponse)
        @self.base_endpoint
        async def register_app(
            email: EmailStr = Body(..., description="User's email address"),
            password: str = Body(..., description="User's password"),
        ):
            """
            Register a new user with the given email and password.
            """
            result = await self.service.register(email, password)
            return result

        @self.router.post(
            "/verify_email", response_model=WrappedGenericMessageResponse
        )
        @self.base_endpoint
        async def verify_email_app(
            email: EmailStr = Body(..., description="User's email address"),
            verification_code: str = Body(
                ..., description="Email verification code"
            ),
        ):
            """
            Verify a user's email address.

            This endpoint is used to confirm a user's email address using the verification code
            sent to their email after registration.
            """
            result = await self.service.verify_email(email, verification_code)
            return GenericMessageResponse(message=result["message"])

        @self.router.post("/login", response_model=WrappedTokenResponse)
        @self.base_endpoint
        async def login_app(
            form_data: OAuth2PasswordRequestForm = Depends(),
        ):
            """
            Authenticate a user and provide access tokens.

            This endpoint authenticates a user using their email (username) and password,
            and returns access and refresh tokens upon successful authentication.
            """
            login_result = await self.service.login(
                form_data.username, form_data.password
            )
            return login_result

        @self.router.post(
            "/logout", response_model=WrappedGenericMessageResponse
        )
        @self.base_endpoint
        async def logout_app(
            token: str = Depends(oauth2_scheme),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
        ):
            """
            Log out the current user.

            This endpoint invalidates the user's current access token, effectively logging them out.
            """
            result = await self.service.logout(token)
            return GenericMessageResponse(message=result["message"])

        @self.router.get("/user", response_model=WrappedUserResponse)
        @self.base_endpoint
        async def get_user_app(
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
        ):
            """
            Get the current user's profile information.

            This endpoint returns the profile information of the currently authenticated user.
            """
            return auth_user

        @self.router.put("/user", response_model=WrappedUserResponse)
        @self.base_endpoint
        async def put_user_app(
            email: EmailStr | None = Body(
                None, description="Updated email address"
            ),
            name: str | None = Body(None, description="Updated user name"),
            bio: str | None = Body(None, description="Updated user bio"),
            profile_picture: str | None = Body(
                None, description="Updated profile picture URL"
            ),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
        ):
            """
            Update the current user's profile information.

            This endpoint allows the authenticated user to update their profile information.
            """
            return await self.service.update_user(
                user_id=auth_user.id,
                email=email,
                name=name,
                bio=bio,
                profile_picture=profile_picture,
            )

        @self.router.post(
            "/refresh_access_token", response_model=WrappedTokenResponse
        )
        @self.base_endpoint
        async def refresh_access_token_app(
            refresh_token: str = Body(..., description="Refresh token")
        ):
            """
            Refresh the access token using a refresh token.

            This endpoint allows users to obtain a new access token using their refresh token.
            """
            refresh_result = await self.service.refresh_access_token(
                refresh_token=refresh_token,
            )
            return refresh_result

        @self.router.post(
            "/change_password", response_model=WrappedGenericMessageResponse
        )
        @self.base_endpoint
        async def change_password_app(
            current_password: str = Body(..., description="Current password"),
            new_password: str = Body(..., description="New password"),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
        ):
            """
            Change the authenticated user's password.

            This endpoint allows users to change their password by providing their current password
            and a new password.
            """
            result = await self.service.change_password(
                auth_user,
                current_password,
                new_password,
            )
            return GenericMessageResponse(message=result["message"])

        @self.router.post(
            "/request_password_reset",
            response_model=WrappedGenericMessageResponse,
        )
        @self.base_endpoint
        async def request_password_reset_app(
            email: EmailStr = Body(..., description="User's email address")
        ):
            """
            Request a password reset for a user.

            This endpoint initiates the password reset process by sending a reset link
            to the specified email address.
            """
            result = await self.service.request_password_reset(email)
            return GenericMessageResponse(message=result["message"])

        @self.router.post(
            "/reset_password",
            response_model=WrappedGenericMessageResponse,
        )
        @self.base_endpoint
        async def reset_password_app(
            reset_token: str = Body(..., description="Password reset token"),
            new_password: str = Body(..., description="New password"),
        ):
            result = await self.service.confirm_password_reset(
                reset_token, new_password
            )
            return GenericMessageResponse(message=result["message"])

        @self.router.delete(
            "/user/{user_id}", response_model=WrappedGenericMessageResponse
        )
        @self.base_endpoint
        async def delete_user_app(
            user_id: str = Path(..., description="ID of the user to delete"),
            password: Optional[str] = Body(
                None, description="User's current password"
            ),
            delete_vector_data: Optional[bool] = Body(
                False,
                description="Whether to delete the user's vector data",
            ),
            auth_user=Depends(self.service.providers.auth.auth_wrapper),
        ):
            """
            Delete a user account.

            This endpoint allows users to delete their own account or, for superusers,
            to delete any user account.
            """
            if auth_user.id != user_id and not auth_user.is_superuser:
                raise Exception("User ID does not match authenticated user")
            if not auth_user.is_superuser and not password:
                raise Exception("Password is required for non-superusers")
            user_uuid = uuid.UUID(user_id)
            result = await self.service.delete_user(
                user_uuid, password, delete_vector_data
            )
            return GenericMessageResponse(message=result["message"])



from typing import Any, Dict

import asyncclick as click
from asyncclick import pass_context

from cli.command_group import cli
from cli.utils.param_types import JSON
from cli.utils.timer import timer


@cli.command()
@click.option("--filters", type=JSON, help="Filters for analytics as JSON")
@click.option("--analysis-types", type=JSON, help="Analysis types as JSON")
@pass_context
def analytics(ctx, filters: Dict[str, Any], analysis_types: Dict[str, Any]):
    client = ctx.obj
    """Retrieve analytics data."""
    with timer():
        response = client.analytics(filters, analysis_types)

    click.echo(response)


@cli.command()
@pass_context
def app_settings(client):
    """Retrieve application settings."""
    with timer():
        response = client.app_settings()

    click.echo(response)


# TODO: Implement score_completion


@cli.command()
@click.option("--user-ids", multiple=True, help="User IDs to overview")
@click.option(
    "--offset",
    default=None,
    help="The offset to start from. Defaults to 0.",
)
@click.option(
    "--limit",
    default=None,
    help="The maximum number of nodes to return. Defaults to 100.",
)
@pass_context
def users_overview(ctx, user_ids, offset, limit):
    """Get an overview of users."""
    client = ctx.obj
    user_ids = list(user_ids) if user_ids else None

    with timer():
        response = client.users_overview(user_ids, offset, limit)

    if "results" in response:
        click.echo("\nUser Overview:")
        click.echo(
            f"{'User ID':<40} {'Num Files':<10} {'Total Size (bytes)':<20} Document IDs"
        )
        for user in response["results"]:
            click.echo(
                f"{user['user_id']:<40} {user['num_files']:<10} {user['total_size_in_bytes']:<20} {', '.join(user['document_ids'][:3]) + ('...' if len(user['document_ids']) > 3 else '')}"
            )
    else:
        click.echo("No users found.")


@cli.command()
@click.option(
    "--filter",
    "-f",
    multiple=True,
    help="Filters for deletion in the format key:operator:value",
)
@pass_context
def delete(ctx, filter):
    """Delete documents based on filters."""
    client = ctx.obj
    filters = {}
    for f in filter:
        key, operator, value = f.split(":", 2)
        if key not in filters:
            filters[key] = {}
        filters[key][f"${operator}"] = value

    with timer():
        response = client.delete(filters=filters)

    click.echo(response)


@cli.command()
@click.option("--document-ids", multiple=True, help="Document IDs to overview")
@click.option(
    "--offset",
    default=None,
    help="The offset to start from. Defaults to 0.",
)
@click.option(
    "--limit",
    default=None,
    help="The maximum number of nodes to return. Defaults to 100.",
)
@pass_context
def documents_overview(ctx, document_ids, offset, limit):
    """Get an overview of documents."""
    client = ctx.obj
    document_ids = list(document_ids) if document_ids else None

    with timer():
        response = client.documents_overview(document_ids, offset, limit)

    for document in response["results"]:
        click.echo(document)


@cli.command()
@click.option("--document-id", help="Document ID to retrieve chunks for")
@click.option(
    "--offset",
    default=None,
    help="The offset to start from. Defaults to 0.",
)
@click.option(
    "--limit",
    default=None,
    help="The maximum number of nodes to return. Defaults to 100.",
)
@pass_context
def document_chunks(ctx, document_id, offset, limit):
    """Get chunks of a specific document."""
    client = ctx.obj
    if not document_id:
        click.echo("Error: Document ID is required.")
        return

    with timer():
        chunks_data = client.document_chunks(document_id, offset, limit)

    chunks = chunks_data["results"]
    if not chunks:
        click.echo("No chunks found for the given document ID.")
        return

    click.echo(f"\nNumber of chunks: {len(chunks)}")

    for index, chunk in enumerate(chunks, 1):
        click.echo(f"\nChunk {index}:")
        if isinstance(chunk, dict):
            click.echo(f"Fragment ID: {chunk.get('fragment_id', 'N/A')}")
            click.echo(f"Text: {chunk.get('text', '')[:100]}...")
            click.echo(f"Metadata: {chunk.get('metadata', {})}")
        else:
            click.echo(f"Unexpected chunk format: {chunk}")


@cli.command()
@click.option(
    "--offset",
    default=None,
    help="The offset to start from. Defaults to 0.",
)
@click.option(
    "--limit",
    default=None,
    help="The maximum number of nodes to return. Defaults to 100.",
)
@pass_context
def inspect_knowledge_graph(ctx, offset, limit):
    """Inspect the knowledge graph."""
    client = ctx.obj
    with timer():
        response = client.inspect_knowledge_graph(offset, limit)

    click.echo(response["results"])



import asyncio
import inspect
import json
from typing import AsyncGenerator, Generator

import httpx
import nest_asyncio
from fastapi.testclient import TestClient

from .auth import AuthMethods
from .ingestion import IngestionMethods
from .management import ManagementMethods
from .models import R2RException
from .restructure import RestructureMethods
from .retrieval import RetrievalMethods
from .server import ServerMethods

nest_asyncio.apply()

# The empty args become necessary after a recent modification to `base_endpoint`
# TODO - Remove the explicitly empty args
EMPTY_ARGS = {"args": "", "kwargs": "{}"}


def handle_request_error(response):
    if response.status_code < 400:
        return

    try:
        error_content = response.json()
        if isinstance(error_content, dict) and "detail" in error_content:
            detail = error_content["detail"]
            if isinstance(detail, dict):
                message = detail.get("message", str(response.text))
            else:
                message = str(detail)
        else:
            message = str(error_content)
    except json.JSONDecodeError:
        message = response.text

    raise R2RException(
        status_code=response.status_code,
        message=message,
    )


async def handle_request_error_async(response):
    if response.status_code < 400:
        return

    try:
        if response.headers.get("content-type") == "application/json":
            error_content = await response.json()
        else:
            error_content = await response.text

        if isinstance(error_content, dict) and "detail" in error_content:
            detail = error_content["detail"]
            if isinstance(detail, dict):
                message = detail.get("message", str(error_content))
            else:
                message = str(detail)
        else:
            message = str(error_content)
    except Exception:
        message = response.text

    raise R2RException(
        status_code=response.status_code,
        message=message,
    )


class R2RAsyncClient:
    def __init__(
        self,
        base_url: str = "http://localhost:7272",
        prefix: str = "/v2",
        custom_client=None,
        timeout: float = 300.0,
    ):
        self.base_url = base_url
        self.prefix = prefix
        self.access_token = None
        self._refresh_token = None
        self.client = custom_client or httpx.AsyncClient(timeout=timeout)
        self.timeout = timeout

        # Initialize methods grouop
        self._auth = AuthMethods
        self._ingestion = IngestionMethods
        self._management = ManagementMethods
        self._restructure = RestructureMethods
        self._retrieval = RetrievalMethods
        self._server = ServerMethods

        # Collect all methods from the methods group
        self._methods = {}
        for collection in [
            self._auth,
            self._ingestion,
            self._management,
            self._restructure,
            self._retrieval,
            self._server,
        ]:
            for name, method in inspect.getmembers(
                collection, predicate=inspect.isfunction
            ):
                if not name.startswith("_"):
                    self._methods[name] = method

    async def _make_request(self, method, endpoint, **kwargs):
        url = f"{self.base_url}{self.prefix}/{endpoint}"
        headers = kwargs.pop("headers", {})
        if self.access_token and endpoint not in [
            "register",
            "login",
            "verify_email",
        ]:
            headers.update(self._get_auth_header())
        if (
            kwargs.get("params", None) == {}
            or kwargs.get("params", None) is None
        ):
            if "params" in kwargs:
                kwargs.pop("params")

        if isinstance(self.client, TestClient):
            # Weird mocking fix...
            params = kwargs.pop("params", {})
            params = {**params, **EMPTY_ARGS}

            response = getattr(self.client, method.lower())(
                url, headers=headers, params=params, **kwargs
            )
            return response.json() if response.content else None
        else:
            try:
                response = await self.client.request(
                    method, url, headers=headers, **kwargs
                )
                await handle_request_error_async(response)
                return response.json() if response.content else None
            except httpx.RequestError as e:
                raise R2RException(
                    status_code=500, message=f"Request failed: {str(e)}"
                ) from e

    async def _make_streaming_request(
        self, method: str, endpoint: str, **kwargs
    ) -> AsyncGenerator[str, None]:
        url = f"{self.base_url}{self.prefix}/{endpoint}"
        headers = kwargs.pop("headers", {})
        if self.access_token and endpoint not in [
            "register",
            "login",
            "verify_email",
        ]:
            headers.update(self._get_auth_header())

        async with httpx.AsyncClient() as client:
            async with client.stream(
                method, url, headers=headers, timeout=self.timeout, **kwargs
            ) as response:
                handle_request_error(response)
                async for chunk in response.aiter_text():
                    yield chunk

    def _get_auth_header(self) -> dict:
        if not self.access_token:
            return {}
        return {"Authorization": f"Bearer {self.access_token}"}

    def _ensure_authenticated(self):
        if not self.access_token:
            raise R2RException(
                status_code=401,
                message="Not authenticated. Please login first.",
            )

    async def close(self):
        await self.client.aclose()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    def __getattr__(self, name):
        if name in self._methods:
            return lambda *args, **kwargs: self._methods[name](
                self, *args, **kwargs
            )
        raise AttributeError(f"'R2RClient' object has no attribute '{name}'")

    def __dir__(self):
        return list(set(super().__dir__() + list(self._methods.keys())))


class R2RClient:
    def __init__(self, *args, **kwargs):
        self.async_client = R2RAsyncClient(*args, **kwargs)

    def _sync_generator(self, async_gen: AsyncGenerator) -> Generator:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        try:
            while True:
                yield loop.run_until_complete(async_gen.__anext__())
        except StopAsyncIteration:
            pass
        finally:
            loop.close()

    def __getattr__(self, name):
        async_attr = getattr(self.async_client, name)
        if callable(async_attr):

            def sync_wrapper(*args, **kwargs):
                result = asyncio.get_event_loop().run_until_complete(
                    async_attr(*args, **kwargs)
                )
                if isinstance(result, AsyncGenerator):
                    return self._sync_generator(result)
                return result

            return sync_wrapper
        return async_attr

    def __dir__(self):
        return dir(self.async_client)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        asyncio.get_event_loop().run_until_complete(self.async_client.close())



from .cli import cli, main
from .command_group import cli as command_group_cli
from .commands import auth, ingestion, management, retrieval, server

__all__ = [
    # From cli.py
    "cli",
    "main",
    # From Command Collection
    "command_group_cli",
    # From Commands
    "auth",
    "ingestion",
    "management",
    "restructure",
    "retrieval",
    "server",
]



import logging
import os
from typing import Any, Optional

from core.agent import R2RRAGAgent, R2RStreamingRAGAgent
from core.base import (
    AsyncPipe,
    AuthConfig,
    AuthProvider,
    ChunkingConfig,
    ChunkingProvider,
    CompletionConfig,
    CompletionProvider,
    CryptoConfig,
    CryptoProvider,
    DatabaseConfig,
    DatabaseProvider,
    EmbeddingConfig,
    EmbeddingProvider,
    FileConfig,
    FileProvider,
    KGProvider,
    ParsingConfig,
    ParsingProvider,
    PromptConfig,
    PromptProvider,
    RunLoggingSingleton,
)
from core.pipelines import RAGPipeline, SearchPipeline
from core.pipes import GeneratorPipe, MultiSearchPipe, SearchPipe

from ..abstractions import R2RAgents, R2RPipelines, R2RPipes, R2RProviders
from ..config import R2RConfig

logger = logging.getLogger(__name__)


class R2RProviderFactory:
    def __init__(self, config: R2RConfig):
        self.config = config

    @staticmethod
    async def create_auth_provider(
        auth_config: AuthConfig,
        db_provider: DatabaseProvider,
        crypto_provider: CryptoProvider,
        *args,
        **kwargs,
    ) -> AuthProvider:
        if auth_config.provider == "r2r":
            from core.providers import R2RAuthProvider

            auth_provider = R2RAuthProvider(
                auth_config, crypto_provider, db_provider
            )
            await auth_provider.initialize()
            return auth_provider
        else:
            raise ValueError(
                f"Auth provider {auth_config.provider} not supported."
            )

    @staticmethod
    def create_crypto_provider(
        crypto_config: CryptoConfig, *args, **kwargs
    ) -> CryptoProvider:
        if crypto_config.provider == "bcrypt":
            from core.providers.crypto import BCryptConfig, BCryptProvider

            return BCryptProvider(BCryptConfig(**crypto_config.dict()))
        else:
            raise ValueError(
                f"Crypto provider {crypto_config.provider} not supported."
            )

    @staticmethod
    def create_parsing_provider(
        parsing_config: ParsingConfig, *args, **kwargs
    ) -> ParsingProvider:
        if parsing_config.provider == "r2r":
            from core.providers import R2RParsingProvider

            return R2RParsingProvider(parsing_config)
        elif parsing_config.provider in [
            "unstructured_local",
            "unstructured_api",
        ]:
            from core.providers import UnstructuredParsingProvider

            return UnstructuredParsingProvider(
                parsing_config.provider == "unstructured_api", parsing_config
            )
        else:
            raise ValueError(
                f"Parsing provider {parsing_config.provider} not supported"
            )

    @staticmethod
    def create_chunking_provider(
        chunking_config: ChunkingConfig, *args, **kwargs
    ) -> ChunkingProvider:
        chunking_config.validate_config()
        if chunking_config.provider == "r2r":
            from core.base import R2RChunkingConfig
            from core.providers import R2RChunkingProvider

            chunking_config_r2r = R2RChunkingConfig(
                **chunking_config.extra_fields
            )
            return R2RChunkingProvider(chunking_config_r2r)
        elif chunking_config.provider in [
            "unstructured_local",
            "unstructured_api",
        ]:
            from core.base import UnstructuredChunkingConfig
            from core.providers import UnstructuredChunkingProvider

            chunking_config_unst = UnstructuredChunkingConfig(
                **chunking_config.extra_fields
            )
            return UnstructuredChunkingProvider(chunking_config_unst)
        else:
            raise ValueError(
                f"Chunking provider {chunking_config.provider} not supported"
            )

    @staticmethod
    def create_orchestration_provider(*args, **kwargs):
        from core.base.providers import OrchestrationConfig
        from core.providers import HatchetOrchestrationProvider

        orchestration_provider = HatchetOrchestrationProvider(
            OrchestrationConfig(provider="hatchet")
        )
        orchestration_provider.get_worker("r2r-worker")
        return orchestration_provider

    async def create_database_provider(
        self,
        db_config: DatabaseConfig,
        crypto_provider: CryptoProvider,
        *args,
        **kwargs,
    ) -> DatabaseProvider:
        database_provider: Optional[DatabaseProvider] = None
        if not self.config.embedding.base_dimension:
            raise ValueError(
                "Embedding config must have a base dimension to initialize database."
            )

        vector_db_dimension = self.config.embedding.base_dimension
        if db_config.provider == "postgres":
            from core.providers import PostgresDBProvider

            database_provider = PostgresDBProvider(
                db_config, vector_db_dimension, crypto_provider=crypto_provider
            )
            await database_provider.initialize()
            return database_provider
        else:
            raise ValueError(
                f"Database provider {db_config.provider} not supported"
            )

    @staticmethod
    def create_embedding_provider(
        embedding: EmbeddingConfig, *args, **kwargs
    ) -> EmbeddingProvider:
        embedding_provider: Optional[EmbeddingProvider] = None

        if embedding.provider == "openai":
            if not os.getenv("OPENAI_API_KEY"):
                raise ValueError(
                    "Must set OPENAI_API_KEY in order to initialize OpenAIEmbeddingProvider."
                )
            from core.providers import OpenAIEmbeddingProvider

            embedding_provider = OpenAIEmbeddingProvider(embedding)

        elif embedding.provider == "litellm":
            from core.providers import LiteLLMEmbeddingProvider

            embedding_provider = LiteLLMEmbeddingProvider(embedding)

        elif embedding.provider == "ollama":
            from core.providers import OllamaEmbeddingProvider

            embedding_provider = OllamaEmbeddingProvider(embedding)

        elif embedding is None:
            embedding_provider = None

        else:
            raise ValueError(
                f"Embedding provider {embedding.provider} not supported"
            )

        return embedding_provider

    @staticmethod
    async def create_file_provider(
        file_config: FileConfig,
        db_provider: Any,
        *args,
        **kwargs,
    ) -> FileProvider:
        file_provider: Optional[FileProvider] = None
        if file_config.provider == "postgres":
            from core.providers import PostgresFileProvider

            file_provider = PostgresFileProvider(file_config, db_provider)
            await file_provider.initialize()
        else:
            raise ValueError(
                f"File provider {file_config.provider} not supported."
            )

        return file_provider

    @staticmethod
    def create_llm_provider(
        llm_config: CompletionConfig, *args, **kwargs
    ) -> CompletionProvider:
        llm_provider: Optional[CompletionProvider] = None
        if llm_config.provider == "openai":
            from core.providers import OpenAICompletionProvider

            llm_provider = OpenAICompletionProvider(llm_config)
        elif llm_config.provider == "litellm":
            from core.providers import LiteCompletionProvider

            llm_provider = LiteCompletionProvider(llm_config)
        else:
            raise ValueError(
                f"Language model provider {llm_config.provider} not supported"
            )
        if not llm_provider:
            raise ValueError("Language model provider not found")
        return llm_provider

    @staticmethod
    async def create_prompt_provider(
        prompt_config: PromptConfig,
        db_provider: DatabaseProvider,
        *args,
        **kwargs,
    ) -> PromptProvider:
        prompt_provider = None

        if prompt_config.provider != "r2r":
            raise ValueError(
                f"Prompt provider {prompt_config.provider} not supported"
            )
        from core.providers import R2RPromptProvider

        prompt_provider = R2RPromptProvider(prompt_config, db_provider)
        await prompt_provider.initialize()

        return prompt_provider

    @staticmethod
    def create_kg_provider(kg_config, *args, **kwargs):
        if kg_config.provider == "neo4j":
            from core.providers import Neo4jKGProvider

            return Neo4jKGProvider(kg_config)
        elif kg_config.provider is None:
            return None
        else:
            raise ValueError(
                f"KG provider {kg_config.provider} not supported."
            )

    async def create_providers(
        self,
        embedding_provider_override: Optional[EmbeddingProvider] = None,
        llm_provider_override: Optional[CompletionProvider] = None,
        prompt_provider_override: Optional[PromptProvider] = None,
        kg_provider_override: Optional[KGProvider] = None,
        crypto_provider_override: Optional[CryptoProvider] = None,
        auth_provider_override: Optional[AuthProvider] = None,
        database_provider_override: Optional[DatabaseProvider] = None,
        parsing_provider_override: Optional[ParsingProvider] = None,
        chunking_config: Optional[ChunkingProvider] = None,
        file_provider_override: Optional[FileProvider] = None,
        orchestration_provider_override: Optional[Any] = None,
        *args,
        **kwargs,
    ) -> R2RProviders:

        embedding_provider = (
            embedding_provider_override
            or self.create_embedding_provider(
                self.config.embedding, *args, **kwargs
            )
        )

        llm_provider = llm_provider_override or self.create_llm_provider(
            self.config.completion, *args, **kwargs
        )

        kg_provider = kg_provider_override or self.create_kg_provider(
            self.config.kg, *args, **kwargs
        )

        crypto_provider = (
            crypto_provider_override
            or self.create_crypto_provider(self.config.crypto, *args, **kwargs)
        )

        database_provider = (
            database_provider_override
            or await self.create_database_provider(
                self.config.database, crypto_provider, *args, **kwargs
            )
        )

        auth_provider = (
            auth_provider_override
            or await self.create_auth_provider(
                self.config.auth,
                database_provider,
                crypto_provider,
                *args,
                **kwargs,
            )
        )

        prompt_provider = (
            prompt_provider_override
            or await self.create_prompt_provider(
                self.config.prompt, database_provider, *args, **kwargs
            )
        )

        parsing_provider = (
            parsing_provider_override
            or self.create_parsing_provider(
                self.config.parsing, *args, **kwargs
            )
        )

        chunking_provider = chunking_config or self.create_chunking_provider(
            self.config.chunking, *args, **kwargs
        )

        file_provider = file_provider_override or await self.create_file_provider(
            self.config.file, database_provider, *args, **kwargs  # type: ignore
        )

        orchestration_provider = (
            orchestration_provider_override
            or self.create_orchestration_provider()
        )

        return R2RProviders(
            auth=auth_provider,
            chunking=chunking_provider,
            database=database_provider,
            embedding=embedding_provider,
            llm=llm_provider,
            parsing=parsing_provider,
            prompt=prompt_provider,
            kg=kg_provider,
            orchestration=orchestration_provider,
            file=file_provider,
        )


class R2RPipeFactory:
    def __init__(self, config: R2RConfig, providers: R2RProviders):
        self.config = config
        self.providers = providers

    def create_pipes(
        self,
        parsing_pipe_override: Optional[AsyncPipe] = None,
        embedding_pipe_override: Optional[AsyncPipe] = None,
        kg_extraction_pipe_override: Optional[AsyncPipe] = None,
        kg_storage_pipe_override: Optional[AsyncPipe] = None,
        kg_search_pipe_override: Optional[AsyncPipe] = None,
        vector_storage_pipe_override: Optional[AsyncPipe] = None,
        vector_search_pipe_override: Optional[AsyncPipe] = None,
        rag_pipe_override: Optional[AsyncPipe] = None,
        streaming_rag_pipe_override: Optional[AsyncPipe] = None,
        kg_node_extraction_pipe: Optional[AsyncPipe] = None,
        kg_node_description_pipe: Optional[AsyncPipe] = None,
        kg_clustering_pipe: Optional[AsyncPipe] = None,
        kg_community_summary_pipe: Optional[AsyncPipe] = None,
        chunking_pipe_override: Optional[AsyncPipe] = None,
        *args,
        **kwargs,
    ) -> R2RPipes:
        return R2RPipes(
            parsing_pipe=parsing_pipe_override
            or self.create_parsing_pipe(
                self.config.parsing.excluded_parsers,
                self.config.parsing.override_parsers,
                *args,
                **kwargs,
            ),
            embedding_pipe=embedding_pipe_override
            or self.create_embedding_pipe(*args, **kwargs),
            kg_extraction_pipe=kg_extraction_pipe_override
            or self.create_kg_extraction_pipe(*args, **kwargs),
            kg_storage_pipe=kg_storage_pipe_override
            or self.create_kg_storage_pipe(*args, **kwargs),
            vector_storage_pipe=vector_storage_pipe_override
            or self.create_vector_storage_pipe(*args, **kwargs),
            vector_search_pipe=vector_search_pipe_override
            or self.create_vector_search_pipe(*args, **kwargs),
            kg_search_pipe=kg_search_pipe_override
            or self.create_kg_search_pipe(*args, **kwargs),
            rag_pipe=rag_pipe_override
            or self.create_rag_pipe(*args, **kwargs),
            streaming_rag_pipe=streaming_rag_pipe_override
            or self.create_rag_pipe(True, *args, **kwargs),
            kg_node_extraction_pipe=kg_node_extraction_pipe
            or self.create_kg_node_extraction_pipe(*args, **kwargs),
            kg_node_description_pipe=kg_node_description_pipe
            or self.create_kg_node_description_pipe(*args, **kwargs),
            kg_clustering_pipe=kg_clustering_pipe
            or self.create_kg_clustering_pipe(*args, **kwargs),
            kg_community_summary_pipe=kg_community_summary_pipe
            or self.create_kg_community_summary_pipe(*args, **kwargs),
            chunking_pipe=chunking_pipe_override
            or self.create_chunking_pipe(*args, **kwargs),
        )

    def create_parsing_pipe(self, *args, **kwargs) -> Any:
        from core.pipes import ParsingPipe

        return ParsingPipe(
            parsing_provider=self.providers.parsing,
            file_provider=self.providers.file,
            config=AsyncPipe.PipeConfig(name="parsing_pipe"),
        )

    def create_chunking_pipe(self, *args, **kwargs) -> Any:
        from core.pipes import ChunkingPipe

        return ChunkingPipe(
            chunking_provider=self.providers.chunking,
            config=AsyncPipe.PipeConfig(name="chunking_pipe"),
        )

    def create_embedding_pipe(self, *args, **kwargs) -> Any:
        if self.config.embedding.provider is None:
            return None

        from core.pipes import EmbeddingPipe

        return EmbeddingPipe(
            embedding_provider=self.providers.embedding,
            database_provider=self.providers.database,
            embedding_batch_size=self.config.embedding.batch_size,
            config=AsyncPipe.PipeConfig(name="embedding_pipe"),
        )

    def create_vector_storage_pipe(self, *args, **kwargs) -> Any:
        if self.config.embedding.provider is None:
            return None

        from core.pipes import VectorStoragePipe

        return VectorStoragePipe(
            database_provider=self.providers.database,
            config=AsyncPipe.PipeConfig(name="vector_storage_pipe"),
        )

    def create_default_vector_search_pipe(self, *args, **kwargs) -> Any:
        if self.config.embedding.provider is None:
            return None

        from core.pipes import VectorSearchPipe

        return VectorSearchPipe(
            database_provider=self.providers.database,
            embedding_provider=self.providers.embedding,
            config=SearchPipe.SearchConfig(name="vector_search_pipe"),
        )

    def create_multi_search_pipe(
        self,
        inner_search_pipe: SearchPipe,
        use_rrf: bool = False,
        expansion_technique: str = "hyde",
        expansion_factor: int = 3,
        *args,
        **kwargs,
    ) -> MultiSearchPipe:
        from core.pipes import QueryTransformPipe

        multi_search_config = MultiSearchPipe.PipeConfig(
            use_rrf=use_rrf, expansion_factor=expansion_factor
        )

        query_transform_pipe = QueryTransformPipe(
            llm_provider=self.providers.llm,
            prompt_provider=self.providers.prompt,
            config=QueryTransformPipe.QueryTransformConfig(
                name="multi_query_transform",
                task_prompt=expansion_technique,
            ),
        )

        return MultiSearchPipe(
            query_transform_pipe=query_transform_pipe,
            inner_search_pipe=inner_search_pipe,
            config=multi_search_config,
        )

    def create_vector_search_pipe(self, *args, **kwargs) -> Any:
        if self.config.embedding.provider is None:
            return None

        vanilla_vector_search_pipe = self.create_default_vector_search_pipe(
            *args, **kwargs
        )
        hyde_search_pipe = self.create_multi_search_pipe(
            vanilla_vector_search_pipe,
            False,
            "hyde",
            *args,
            **kwargs,
        )
        rag_fusion_pipe = self.create_multi_search_pipe(
            vanilla_vector_search_pipe,
            True,
            "rag_fusion",
            *args,
            **kwargs,
        )

        from core.pipes import RoutingSearchPipe

        return RoutingSearchPipe(
            search_pipes={
                "vanilla": vanilla_vector_search_pipe,
                "hyde": hyde_search_pipe,
                "rag_fusion": rag_fusion_pipe,
            },
            default_strategy="hyde",
            config=AsyncPipe.PipeConfig(name="routing_search_pipe"),
        )

    def create_kg_extraction_pipe(self, *args, **kwargs) -> Any:
        if self.config.kg.provider is None:
            return None

        from core.pipes import KGTriplesExtractionPipe

        return KGTriplesExtractionPipe(
            kg_provider=self.providers.kg,
            llm_provider=self.providers.llm,
            database_provider=self.providers.database,
            prompt_provider=self.providers.prompt,
            chunking_provider=self.providers.chunking,
            config=AsyncPipe.PipeConfig(name="kg_extraction_pipe"),
        )

    def create_kg_storage_pipe(self, *args, **kwargs) -> Any:
        if self.config.kg.provider is None:
            return None

        from core.pipes import KGStoragePipe

        return KGStoragePipe(
            kg_provider=self.providers.kg,
            embedding_provider=self.providers.embedding,
            config=AsyncPipe.PipeConfig(name="kg_storage_pipe"),
        )

    def create_kg_search_pipe(self, *args, **kwargs) -> Any:
        if self.config.kg.provider is None:
            return None

        from core.pipes import KGSearchSearchPipe

        return KGSearchSearchPipe(
            kg_provider=self.providers.kg,
            llm_provider=self.providers.llm,
            prompt_provider=self.providers.prompt,
            embedding_provider=self.providers.embedding,
            config=GeneratorPipe.PipeConfig(
                name="kg_rag_pipe", task_prompt="kg_search"
            ),
        )

    def create_rag_pipe(self, stream: bool = False, *args, **kwargs) -> Any:
        if stream:
            from core.pipes import StreamingSearchRAGPipe

            return StreamingSearchRAGPipe(
                llm_provider=self.providers.llm,
                prompt_provider=self.providers.prompt,
                config=GeneratorPipe.PipeConfig(
                    name="streaming_rag_pipe", task_prompt="default_rag"
                ),
            )
        else:
            from core.pipes import SearchRAGPipe

            return SearchRAGPipe(
                llm_provider=self.providers.llm,
                prompt_provider=self.providers.prompt,
                config=GeneratorPipe.PipeConfig(
                    name="search_rag_pipe", task_prompt="default_rag"
                ),
            )

    def create_kg_node_extraction_pipe(self, *args, **kwargs) -> Any:
        from core.pipes import KGNodeExtractionPipe

        return KGNodeExtractionPipe(
            kg_provider=self.providers.kg,
            llm_provider=self.providers.llm,
            prompt_provider=self.providers.prompt,
            config=AsyncPipe.PipeConfig(name="kg_node_extraction_pipe"),
        )

    def create_kg_node_description_pipe(self, *args, **kwargs) -> Any:
        from core.pipes import KGNodeDescriptionPipe

        return KGNodeDescriptionPipe(
            kg_provider=self.providers.kg,
            llm_provider=self.providers.llm,
            prompt_provider=self.providers.prompt,
            embedding_provider=self.providers.embedding,
            config=AsyncPipe.PipeConfig(name="kg_node_description_pipe"),
        )

    def create_kg_clustering_pipe(self, *args, **kwargs) -> Any:
        from core.pipes import KGClusteringPipe

        return KGClusteringPipe(
            kg_provider=self.providers.kg,
            llm_provider=self.providers.llm,
            prompt_provider=self.providers.prompt,
            embedding_provider=self.providers.embedding,
            config=AsyncPipe.PipeConfig(name="kg_clustering_pipe"),
        )

    def create_kg_community_summary_pipe(self, *args, **kwargs) -> Any:
        from core.pipes import KGCommunitySummaryPipe

        return KGCommunitySummaryPipe(
            kg_provider=self.providers.kg,
            llm_provider=self.providers.llm,
            prompt_provider=self.providers.prompt,
            embedding_provider=self.providers.embedding,
            config=AsyncPipe.PipeConfig(name="kg_community_summary_pipe"),
        )


class R2RPipelineFactory:
    def __init__(self, config: R2RConfig, pipes: R2RPipes):
        self.config = config
        self.pipes = pipes

    def create_search_pipeline(self, *args, **kwargs) -> SearchPipeline:
        """factory method to create an ingestion pipeline."""
        search_pipeline = SearchPipeline()

        # Add vector search pipes if embedding provider and vector provider is set
        if (
            self.config.embedding.provider is not None
            and self.config.database.provider is not None
        ):
            search_pipeline.add_pipe(
                self.pipes.vector_search_pipe, vector_search_pipe=True
            )

        # Add KG pipes if provider is set
        if self.config.kg.provider is not None:
            search_pipeline.add_pipe(
                self.pipes.kg_search_pipe, kg_extraction_pipe=True
            )

        return search_pipeline

    def create_rag_pipeline(
        self,
        search_pipeline: SearchPipeline,
        stream: bool = False,
        *args,
        **kwargs,
    ) -> RAGPipeline:
        rag_pipe = (
            self.pipes.streaming_rag_pipe if stream else self.pipes.rag_pipe
        )

        rag_pipeline = RAGPipeline()
        rag_pipeline.set_search_pipeline(search_pipeline)
        rag_pipeline.add_pipe(rag_pipe)
        return rag_pipeline

    def create_pipelines(
        self,
        search_pipeline: Optional[SearchPipeline] = None,
        rag_pipeline: Optional[RAGPipeline] = None,
        streaming_rag_pipeline: Optional[RAGPipeline] = None,
        *args,
        **kwargs,
    ) -> R2RPipelines:
        try:
            self.configure_logging()
        except Exception as e:
            logger.warning(f"Error configuring logging: {e}")
        search_pipeline = search_pipeline or self.create_search_pipeline(
            *args, **kwargs
        )
        return R2RPipelines(
            search_pipeline=search_pipeline,
            rag_pipeline=rag_pipeline
            or self.create_rag_pipeline(
                search_pipeline,
                False,
                *args,
                **kwargs,
            ),
            streaming_rag_pipeline=streaming_rag_pipeline
            or self.create_rag_pipeline(
                search_pipeline,
                True,
                *args,
                **kwargs,
            ),
        )

    def configure_logging(self):
        RunLoggingSingleton.configure(self.config.logging)


class R2RAgentFactory:
    def __init__(
        self,
        config: R2RConfig,
        providers: R2RProviders,
        pipelines: R2RPipelines,
    ):
        self.config = config
        self.providers = providers
        self.pipelines = pipelines

    def create_agents(
        self,
        rag_agent_override: Optional[R2RRAGAgent] = None,
        stream_rag_agent_override: Optional[R2RStreamingRAGAgent] = None,
        *args,
        **kwargs,
    ) -> R2RAgents:
        return R2RAgents(
            rag_agent=rag_agent_override
            or self.create_rag_agent(*args, **kwargs),
            streaming_rag_agent=stream_rag_agent_override
            or self.create_streaming_rag_agent(*args, **kwargs),
        )

    def create_streaming_rag_agent(
        self, *args, **kwargs
    ) -> R2RStreamingRAGAgent:
        if not self.providers.llm or not self.providers.prompt:
            raise ValueError(
                "LLM and Prompt providers are required for RAG Agent"
            )

        return R2RStreamingRAGAgent(
            llm_provider=self.providers.llm,
            prompt_provider=self.providers.prompt,
            config=self.config.agent,
            search_pipeline=self.pipelines.search_pipeline,
        )

    def create_rag_agent(self, *args, **kwargs) -> R2RRAGAgent:
        if not self.providers.llm or not self.providers.prompt:
            raise ValueError(
                "LLM and Prompt providers are required for RAG Agent"
            )
        return R2RRAGAgent(
            llm_provider=self.providers.llm,
            prompt_provider=self.providers.prompt,
            config=self.config.agent,
            search_pipeline=self.pipelines.search_pipeline,
        )



import logging

# Keep '*' imports for enhanced development velocity
# corresponding flake8 error codes are F403, F405
from .agent import *
from .base import *
from .integrations import *
from .main import *
from .parsers import *
from .pipelines import *
from .pipes import *
from .providers import *

logger = logging.getLogger("core")
logger.setLevel(logging.INFO)

# Create a console handler and set the level to info
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)

# Create a formatter and set it for the handler
formatter = logging.Formatter(
    "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
)
ch.setFormatter(formatter)

# Add the handler to the logger
logger.addHandler(ch)

# Optional: Prevent propagation to the root logger
logger.propagate = False


__all__ = [
    ## AGENT
    # Base
    "R2RAgent",
    "R2RStreamingAgent",
    # RAG Agents
    "R2RRAGAgent",
    "R2RStreamingRAGAgent",
    ## BASE
    # Base abstractions
    "AsyncSyncMeta",
    "syncable",
    # Completion abstractions
    "CompletionRecord",
    "MessageType",
    # Document abstractions
    "DataType",
    "Document",
    "DocumentExtraction",
    "DocumentFragment",
    "DocumentInfo",
    "IngestionStatus",
    "RestructureStatus",
    "DocumentType",
    # Embedding abstractions
    "EmbeddingPurpose",
    "default_embedding_prefixes",
    # Exception abstractions
    "R2RDocumentProcessingError",
    "R2RException",
    # KG abstractions
    "Entity",
    "KGExtraction",
    "Triple",
    # LLM abstractions
    "GenerationConfig",
    "LLMChatCompletion",
    "LLMChatCompletionChunk",
    "RAGCompletion",
    # Prompt abstractions
    "Prompt",
    # Search abstractions
    "AggregateSearchResult",
    "KGSearchResult",
    "KGSearchSettings",
    "VectorSearchResult",
    "VectorSearchSettings",
    # User abstractions
    "Token",
    "TokenData",
    "UserStats",
    # Vector abstractions
    "Vector",
    "VectorEntry",
    "VectorType",
    ## AGENT
    # Agent abstractions
    "Agent",
    "AgentConfig",
    "Conversation",
    "Message",
    "Tool",
    "ToolResult",
    ## API
    # Auth Responses
    "GenericMessageResponse",
    "TokenResponse",
    "UserResponse",
    ## LOGGING
    # Basic types
    "RunType",
    "AnalysisTypes",
    "LogAnalytics",
    "LogAnalyticsConfig",
    "LogFilterCriteria",
    "LogProcessor",
    # Logging Providers
    "LocalRunLoggingProvider",
    "LoggingConfig",
    "PostgresLoggingConfig",
    "PostgresRunLoggingProvider",
    "RunLoggingSingleton",
    # Run Manager
    "RunManager",
    "manage_run",
    ## PARSERS
    # Base parser
    "AsyncParser",
    ## PIPELINE
    # Base pipeline
    "AsyncPipeline",
    ## PIPES
    "AsyncPipe",
    "AsyncState",
    "PipeType",
    ## PROVIDERS
    # Base provider classes
    "Provider",
    "ProviderConfig",
    # Auth provider
    "AuthConfig",
    "AuthProvider",
    # Chunking provider
    "UnstructuredChunkingConfig",
    "R2RChunkingConfig",
    "ChunkingConfig",
    "ChunkingProvider",
    "Strategy",
    # Crypto provider
    "CryptoConfig",
    "CryptoProvider",
    # Database providers
    "DatabaseConfig",
    "DatabaseProvider",
    "RelationalDBProvider",
    "VectorDBProvider",
    # Embedding provider
    "EmbeddingConfig",
    "EmbeddingProvider",
    # Knowledge Graph provider
    "KGConfig",
    "KGProvider",
    # LLM provider
    "CompletionConfig",
    "CompletionProvider",
    # Parsing provider
    "ParsingConfig",
    "ParsingProvider",
    "OverrideParser",
    # Prompt provider
    "PromptConfig",
    "PromptProvider",
    ## UTILS
    "RecursiveCharacterTextSplitter",
    "TextSplitter",
    "run_pipeline",
    "to_async_generator",
    "generate_run_id",
    "generate_id_from_label",
    "increment_version",
    "EntityType",
    "RelationshipType",
    "format_entity_types",
    "format_relations",
    ## INTEGRATIONS
    # Serper
    "SerperClient",
    ## MAIN
    ## R2R ABSTRACTIONS
    "R2RProviders",
    "R2RPipes",
    "R2RPipelines",
    "R2RAgents",
    ## R2R APP
    "R2RApp",
    ## R2R APP ENTRY
    # "r2r_app",
    ## R2R ASSEMBLY
    # Builder
    "R2RBuilder",
    # Config
    "R2RConfig",
    # Factory
    "R2RProviderFactory",
    "R2RPipeFactory",
    "R2RPipelineFactory",
    "R2RAgentFactory",
    # R2R Routers
    "AuthRouter",
    "IngestionRouter",
    "ManagementRouter",
    "RetrievalRouter",
    "RestructureRouter",
    ## R2R SERVICES
    "AuthService",
    "IngestionService",
    "ManagementService",
    "RetrievalService",
    "RestructureService",
    ## PARSERS
    # Media parsers
    "AudioParser",
    "DOCXParser",
    "ImageParser",
    "MovieParser",
    "PDFParser",
    "PDFParserUnstructured",
    "PDFParserMarker",
    "PPTParser",
    # Structured parsers
    "CSVParser",
    "CSVParserAdvanced",
    "JSONParser",
    "XLSXParser",
    "XLSXParserAdvanced",
    # Text parsers
    "MDParser",
    "HTMLParser",
    "TextParser",
    ## PIPELINES
    "SearchPipeline",
    "RAGPipeline",
    ## PIPES
    "SearchPipe",
    "EmbeddingPipe",
    "KGTriplesExtractionPipe",
    "ParsingPipe",
    "ChunkingPipe",
    "QueryTransformPipe",
    "SearchRAGPipe",
    "StreamingSearchRAGPipe",
    "VectorSearchPipe",
    "VectorStoragePipe",
    "KGStoragePipe",
    "MultiSearchPipe",
    ## PROVIDERS
    # Auth
    "R2RAuthProvider",
    # Chunking
    "R2RChunkingProvider",
    "UnstructuredChunkingProvider",
    # Crypto
    "BCryptProvider",
    "BCryptConfig",
    # Database
    "PostgresDBProvider",
    # Embeddings
    "LiteLLMEmbeddingProvider",
    "OllamaEmbeddingProvider",
    "OpenAIEmbeddingProvider",
    # KG
    "Neo4jKGProvider",
    # LLM
    "OpenAICompletionProvider",
    "LiteCompletionProvider",
    # Parsing
    "R2RParsingProvider",
    "UnstructuredParsingProvider",
    # Prompts
    "R2RPromptProvider",
]



from .base import r2r_hatchet
from .ingestion_workflow import IngestFilesWorkflow, UpdateFilesWorkflow
from .restructure_workflow import (
    CreateGraphWorkflow,
    EnrichGraphWorkflow,
    KGCommunitySummaryWorkflow,
    KgExtractAndStoreWorkflow,
)

__all__ = [
    "r2r_hatchet",
    "IngestFilesWorkflow",
    "UpdateFilesWorkflow",
    "EnrichGraphWorkflow",
    "CreateGraphWorkflow",
    "KgExtractAndStoreWorkflow",
    "KGCommunitySummaryWorkflow",
]



import asyncio
import json
import logging
import re
import uuid
from typing import Any, AsyncGenerator, Optional, Union

from core.base import (
    AsyncState,
    ChunkingProvider,
    CompletionProvider,
    DatabaseProvider,
    DocumentFragment,
    Entity,
    GenerationConfig,
    KGExtraction,
    KGProvider,
    PipeType,
    PromptProvider,
    R2RDocumentProcessingError,
    R2RException,
    RunLoggingSingleton,
    Triple,
)
from core.base.pipes.base_pipe import AsyncPipe

logger = logging.getLogger(__name__)


MIN_VALID_KG_EXTRACTION_RESPONSE_LENGTH = 128


class ClientError(Exception):
    """Base class for client connection errors."""

    pass


class KGTriplesExtractionPipe(
    AsyncPipe[Union[KGExtraction, R2RDocumentProcessingError]]
):
    """
    Extracts knowledge graph information from document extractions.
    """

    class Input(AsyncPipe.Input):
        message: dict

    def __init__(
        self,
        kg_provider: KGProvider,
        database_provider: DatabaseProvider,
        llm_provider: CompletionProvider,
        prompt_provider: PromptProvider,
        chunking_provider: ChunkingProvider,
        config: AsyncPipe.PipeConfig,
        kg_batch_size: int = 1,
        graph_rag: bool = True,
        id_prefix: str = "demo",
        pipe_logger: Optional[RunLoggingSingleton] = None,
        type: PipeType = PipeType.INGESTOR,
        *args,
        **kwargs,
    ):
        super().__init__(
            pipe_logger=pipe_logger,
            type=type,
            config=config
            or AsyncPipe.PipeConfig(name="default_kg_extraction_pipe"),
        )
        self.kg_provider = kg_provider
        self.prompt_provider = prompt_provider
        self.database_provider = database_provider
        self.llm_provider = llm_provider
        self.chunking_provider = chunking_provider
        self.kg_batch_size = kg_batch_size
        self.id_prefix = id_prefix
        self.pipe_run_info = None
        self.graph_rag = graph_rag

    async def extract_kg(
        self,
        fragments: list[DocumentFragment],
        generation_config: GenerationConfig,
        max_knowledge_triples: int,
        entity_types: list[str],
        relation_types: list[str],
        retries: int = 5,
        delay: int = 2,
    ) -> KGExtraction:
        """
        Extracts NER triples from a fragment with retries.
        """

        # combine all fragments into a single string
        combined_fragment: str = " ".join([fragment.data for fragment in fragments])  # type: ignore

        messages = self.prompt_provider._get_message_payload(
            task_prompt_name=self.kg_provider.config.kg_creation_settings.kg_extraction_prompt,
            task_inputs={
                "input": combined_fragment,
                "max_knowledge_triples": max_knowledge_triples,
                "entity_types": "\n".join(entity_types),
                "relation_types": "\n".join(relation_types),
            },
        )

        for attempt in range(retries):
            try:
                response = await self.llm_provider.aget_completion(
                    messages,
                    generation_config=generation_config,
                )

                kg_extraction = response.choices[0].message.content

                if not kg_extraction:
                    raise R2RException(
                        "No knowledge graph extraction found in the response string, the selected LLM likely failed to format it's response correctly.",
                        400,
                    )

                entity_pattern = (
                    r'\("entity"\${4}([^$]+)\${4}([^$]+)\${4}([^$]+)\)'
                )
                relationship_pattern = r'\("relationship"\${4}([^$]+)\${4}([^$]+)\${4}([^$]+)\${4}([^$]+)\${4}(\d+(?:\.\d+)?)\)'

                def parse_fn(response_str: str) -> Any:
                    entities = re.findall(entity_pattern, response_str)

                    if (
                        len(kg_extraction)
                        > MIN_VALID_KG_EXTRACTION_RESPONSE_LENGTH
                        and len(entities) == 0
                    ):
                        raise R2RException(
                            f"No entities found in the response string, the selected LLM likely failed to format it's response correctly. {response_str}",
                            400,
                        )
                        # logger.warning(
                        #     f"No entities found in the response string, the selected LLM likely failed to format it's response correctly. {response_str}",
                        # )

                    relationships = re.findall(
                        relationship_pattern, response_str
                    )

                    entities_dict = {}
                    for entity in entities:
                        entity_value = entity[0]
                        entity_category = entity[1]
                        entity_description = entity[2]
                        entities_dict[entity_value] = Entity(
                            category=entity_category,
                            description=entity_description,
                            name=entity_value,
                            document_ids=[str(fragments[0].document_id)],
                            text_unit_ids=[
                                str(fragment.id) for fragment in fragments
                            ],
                            attributes={"fragment_text": combined_fragment},
                        )

                    relations_arr = []
                    for relationship in relationships:
                        subject = relationship[0]
                        object = relationship[1]
                        predicate = relationship[2]
                        description = relationship[3]
                        weight = float(relationship[4])

                        # check if subject and object are in entities_dict
                        relations_arr.append(
                            Triple(
                                subject=subject,
                                predicate=predicate,
                                object=object,
                                description=description,
                                weight=weight,
                                document_ids=[str(fragments[0].document_id)],
                                text_unit_ids=[
                                    str(fragment.id) for fragment in fragments
                                ],
                                attributes={
                                    "fragment_text": combined_fragment
                                },
                            )
                        )

                    return entities_dict, relations_arr

                entities, triples = parse_fn(kg_extraction)
                return KGExtraction(
                    fragment_ids=[fragment.id for fragment in fragments],
                    document_id=fragments[0].document_id,
                    entities=entities,
                    triples=triples,
                )

            except (
                ClientError,
                json.JSONDecodeError,
                KeyError,
                IndexError,
                R2RException,
            ) as e:
                if attempt < retries - 1:
                    await asyncio.sleep(delay)
                else:
                    logger.error(
                        f"Failed after retries with for fragment {fragments[0].id} of document {fragments[0].document_id}: {e}"
                    )
                    # raise e # you should raise an error.
        # add metadata to entities and triples

        return KGExtraction(
            fragment_ids=[fragment.id for fragment in fragments],
            document_id=fragments[0].document_id,
            entities={},
            triples=[],
        )

    async def _run_logic(  # type: ignore
        self,
        input: Input,
        state: AsyncState,
        run_id: Any,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[Union[KGExtraction, R2RDocumentProcessingError], None]:

        logger.info("Running KG Extraction Pipe")

        document_id = input.message["document_id"]
        generation_config = input.message["generation_config"]
        fragment_merge_count = input.message["fragment_merge_count"]
        max_knowledge_triples = input.message["max_knowledge_triples"]
        entity_types = input.message["entity_types"]
        relation_types = input.message["relation_types"]

        fragments = [
            DocumentFragment(
                id=extraction["fragment_id"],
                extraction_id=extraction["extraction_id"],
                document_id=extraction["document_id"],
                user_id=extraction["user_id"],
                collection_ids=extraction["collection_ids"],
                data=extraction["text"],
                metadata=extraction["metadata"],
            )
            for extraction in self.database_provider.vector.get_document_chunks(
                document_id=document_id
            )[
                "results"
            ]
        ]

        # sort the fragments accroding to chunk_order field in metadata in ascending order
        fragments = sorted(fragments, key=lambda x: x.metadata["chunk_order"])

        # group these extractions into groups of fragment_merge_count
        fragments_groups = [
            fragments[i : i + fragment_merge_count]
            for i in range(0, len(fragments), fragment_merge_count)
        ]

        logger.info(
            f"Extracting KG Triples from {len(fragments_groups)} fragment groups from originally {len(fragments)} fragments for document {document_id}"
        )

        tasks = [
            asyncio.create_task(
                self.extract_kg(
                    fragments=fragments_group,
                    generation_config=generation_config,
                    max_knowledge_triples=max_knowledge_triples,
                    entity_types=entity_types,
                    relation_types=relation_types,
                )
            )
            for fragments_group in fragments_groups
        ]

        for completed_task in asyncio.as_completed(tasks):
            try:
                yield await completed_task
            except Exception as e:
                logger.error(f"Error in Extracting KG Triples: {e}")
                yield R2RDocumentProcessingError(
                    document_id=document_id,
                    error_message=str(e),
                )



import json
import logging
import os
from pathlib import Path
from typing import Any, Optional

import asyncpg
import yaml

from core.base import DatabaseProvider, Prompt, PromptConfig, PromptProvider
from core.base.utils import generate_id_from_label

logger = logging.getLogger(__name__)


class R2RPromptProvider(PromptProvider):
    def __init__(self, config: PromptConfig, db_provider: DatabaseProvider):
        super().__init__(config)
        self.prompts: dict[str, Prompt] = {}
        self.config: PromptConfig = config
        self.db_provider = db_provider
        self.pool: Optional[asyncpg.pool.Pool] = None  # Initialize pool

    async def __aenter__(self):
        await self.initialize()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._close_connection()

    async def _close_connection(self):
        if self.pool:
            await self.pool.close()
            self.pool = None

    async def initialize(self):
        try:
            self.pool = await asyncpg.create_pool(
                self.db_provider.connection_string
            )
            logger.info(
                "R2RPromptProvider successfully connected to Postgres database."
            )

            async with self.pool.acquire() as conn:
                await conn.execute('CREATE EXTENSION IF NOT EXISTS "lo";')

            await self.create_table()
            await self._load_prompts_from_database()
            await self._load_prompts_from_yaml_directory()
        except Exception as e:
            logger.error(f"Failed to initialize R2RPromptProvider: {e}")
            raise

    def _get_table_name(self, base_name: str) -> str:
        return base_name

    async def create_table(self):
        query = f"""
        CREATE TABLE IF NOT EXISTS {self._get_table_name('prompts')} (
            prompt_id UUID PRIMARY KEY,
            name TEXT NOT NULL UNIQUE,
            template TEXT NOT NULL,
            input_types JSONB NOT NULL,
            created_at TIMESTAMPTZ DEFAULT NOW(),
            updated_at TIMESTAMPTZ DEFAULT NOW()
        );
        """
        try:
            await self.execute_query(query)
            logger.debug("Prompts table ensured in the database.")
        except Exception as e:
            logger.error(f"Failed to create table: {e}")
            raise

    async def execute_query(
        self, query: str, params: Optional[list[Any]] = None
    ) -> Any:
        if not self.pool:
            raise ConnectionError("Database pool is not initialized.")
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                if params:
                    return await conn.execute(query, *params)
                return await conn.execute(query)

    async def fetch_query(
        self, query: str, params: Optional[list[Any]] = None
    ) -> Any:
        if not self.pool:
            raise ConnectionError("Database pool is not initialized.")
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                return (
                    await conn.fetch(query, *params)
                    if params
                    else await conn.fetch(query)
                )

    # FIXME: We really should be taking advantage of Pydantic models here
    # so that we don't have to json.dumps/loads all the time
    async def _load_prompts_from_database(self):
        query = """
        SELECT prompt_id, name, template, input_types, created_at, updated_at
        FROM prompts
        """
        try:
            results = await self.fetch_query(query)
            for row in results:
                (
                    _,
                    name,
                    template,
                    input_types,
                    created_at,
                    updated_at,
                ) = (
                    row["prompt_id"],
                    row["name"],
                    row["template"],
                    json.loads(row["input_types"]),
                    row["created_at"],
                    row["updated_at"],
                )
                self.prompts[name] = Prompt(
                    name=name,
                    template=template,
                    input_types=input_types,
                    created_at=created_at,
                    updated_at=updated_at,
                )
            logger.debug(
                f"Loaded {len(self.prompts)} prompts from the database."
            )
        except Exception as e:
            logger.error(f"Failed to load prompts from database: {e}")
            raise

    async def _load_prompts_from_yaml_directory(
        self, directory_path: Optional[Path] = None
    ):
        if not directory_path:
            directory_path = (
                self.config.file_path
                or Path(os.path.dirname(__file__)) / "defaults"
            )

        if not directory_path.is_dir():
            error_msg = (
                f"The specified path is not a directory: {directory_path}"
            )
            logger.error(error_msg)
            raise ValueError(error_msg)

        logger.info(f"Loading prompts from {directory_path}")
        for yaml_file in directory_path.glob("*.yaml"):
            logger.debug(f"Loading prompts from {yaml_file}")
            try:
                with open(yaml_file, "r") as file:
                    data = yaml.safe_load(file)
                    if not isinstance(data, dict):
                        raise ValueError(
                            f"Invalid format in YAML file {yaml_file}"
                        )
                    for name, prompt_data in data.items():
                        if name not in self.prompts:
                            modify_prompt = True
                        else:
                            modify_prompt = (
                                self.prompts[name].created_at
                                == self.prompts[name].updated_at
                            )

                        if modify_prompt:
                            await self.add_prompt(
                                name,
                                prompt_data["template"],
                                prompt_data.get("input_types", {}),
                                modify_created_at=True,
                            )
            except yaml.YAMLError as e:
                error_msg = (
                    f"Error loading prompts from YAML file {yaml_file}: {e}"
                )
                logger.error(error_msg)
                raise ValueError(error_msg)
            except KeyError as e:
                error_msg = f"Missing key in YAML file {yaml_file}: {e}"
                logger.error(error_msg)
                raise ValueError(error_msg)
            except Exception as e:
                error_msg = (
                    f"Unexpected error loading YAML file {yaml_file}: {e}"
                )
                logger.error(error_msg)
                raise ValueError(error_msg)

    async def add_prompt(
        self,
        name: str,
        template: str,
        input_types: dict[str, str],
        modify_created_at: bool = False,
    ) -> None:
        prompt = Prompt(
            prompt_id=generate_id_from_label(name),
            name=name,
            template=template,
            input_types=input_types,
        )
        self.prompts[name] = prompt
        try:
            await self._save_prompt_to_database(
                prompt, modify_created_at=modify_created_at
            )
            logger.debug(f"Added/Updated prompt '{name}' successfully.")
        except Exception as e:
            logger.error(f"Failed to add/update prompt '{name}': {e}")
            raise

    def get_prompt(
        self,
        prompt_name: str,
        inputs: Optional[dict[str, Any]] = None,
        prompt_override: Optional[str] = None,
    ) -> str:
        if prompt_name not in self.prompts:
            raise ValueError(f"Prompt '{prompt_name}' not found.")
        existing_types = self.prompts[prompt_name].input_types
        prompt = (
            Prompt(
                name=prompt_name,
                template=prompt_override,
                input_types=existing_types,
            )
            if prompt_override
            else self.prompts[prompt_name]
        )
        return (
            prompt.template if inputs is None else prompt.format_prompt(inputs)
        )

    async def update_prompt(
        self,
        name: str,
        template: Optional[str] = None,
        input_types: Optional[dict[str, str]] = None,
    ) -> None:
        if name not in self.prompts:
            raise ValueError(f"Prompt '{name}' not found.")
        prompt = self.prompts[name]
        if template:
            prompt.template = template
        if input_types:
            prompt.input_types = input_types
        try:
            await self._save_prompt_to_database(prompt)
            logger.info(f"Updated prompt '{name}' successfully.")
        except Exception as e:
            logger.error(f"Failed to update prompt '{name}': {e}")
            raise

    def get_all_prompts(self) -> dict[str, Prompt]:
        return self.prompts

    async def delete_prompt(self, name: str) -> None:
        if name not in self.prompts:
            raise ValueError(f"Prompt '{name}' not found.")
        del self.prompts[name]
        query = f"""
        DELETE FROM {self._get_table_name('prompts')}
        WHERE name = $1
        """
        try:
            await self.execute_query(query, [name])
            logger.info(f"Deleted prompt '{name}' successfully.")
        except Exception as e:
            logger.error(f"Failed to delete prompt '{name}': {e}")
            raise

    async def _save_prompt_to_database(
        self, prompt: Prompt, modify_created_at: bool = False
    ):
        try:
            created_at_clause = (
                "created_at = NOW()," if modify_created_at else ""
            )

            query = f"""
            INSERT INTO {self._get_table_name('prompts')}
            (prompt_id, name, template, input_types)
            VALUES ($1, $2, $3, $4)
            ON CONFLICT (name) DO UPDATE SET
                template = EXCLUDED.template,
                input_types = EXCLUDED.input_types,
                {created_at_clause}
                updated_at = NOW();
            """
            await self.execute_query(
                query,
                [
                    generate_id_from_label(prompt.name),
                    prompt.name,
                    prompt.template,
                    json.dumps(prompt.input_types),
                ],
            )
        except Exception as e:
            logger.error(
                f"Failed to save prompt '{prompt.name}' to database: {e}"
            )
            raise



import asyncio
import json
import logging
from typing import Any, AsyncGenerator, Optional
from uuid import UUID

from core.base import (
    AsyncState,
    CompletionProvider,
    EmbeddingProvider,
    KGProvider,
    PipeType,
    PromptProvider,
    RunLoggingSingleton,
)
from core.base.abstractions import (
    KGCommunityResult,
    KGEntityResult,
    KGGlobalResult,
    KGRelationshipResult,
    KGSearchMethod,
    KGSearchResult,
    KGSearchResultType,
    KGSearchSettings,
    R2RException,
)

from ..abstractions.generator_pipe import GeneratorPipe

logger = logging.getLogger(__name__)


class KGSearchSearchPipe(GeneratorPipe):
    """
    Embeds and stores documents using a specified embedding model and database.
    """

    def __init__(
        self,
        kg_provider: KGProvider,
        llm_provider: CompletionProvider,
        prompt_provider: PromptProvider,
        embedding_provider: EmbeddingProvider,
        config: GeneratorPipe.PipeConfig,
        pipe_logger: Optional[RunLoggingSingleton] = None,
        type: PipeType = PipeType.INGESTOR,
        *args,
        **kwargs,
    ):
        """
        Initializes the embedding pipe with necessary components and configurations.
        """
        super().__init__(
            llm_provider,
            prompt_provider,
            config,
            type,
            pipe_logger,
            *args,
            **kwargs,
        )
        self.kg_provider = kg_provider
        self.llm_provider = llm_provider
        self.prompt_provider = prompt_provider
        self.embedding_provider = embedding_provider
        self.pipe_run_info = None

    def filter_responses(self, map_responses):
        filtered_responses = []
        for response in map_responses:
            try:
                parsed_response = json.loads(response)
                for item in parsed_response["points"]:
                    try:
                        if item["score"] > 0:
                            filtered_responses.append(item)
                    except KeyError:
                        # Skip this item if it doesn't have a 'score' key
                        logger.warning(f"Item in response missing 'score' key")
                        continue
            except json.JSONDecodeError:
                logger.warning(
                    f"Response is not valid JSON: {response[:100]}..."
                )
                continue
            except KeyError:
                logger.warning(
                    f"Response is missing 'points' key: {response[:100]}..."
                )
                continue

        filtered_responses = sorted(
            filtered_responses, key=lambda x: x["score"], reverse=True
        )

        responses = "\n".join(
            [
                response.get("description", "")
                for response in filtered_responses
            ]
        )
        return responses

    async def local_search(
        self,
        input: GeneratorPipe.Input,
        state: AsyncState,
        run_id: UUID,
        kg_search_settings: KGSearchSettings,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[KGSearchResult, None]:
        # search over communities and
        # do 3 searches. One over entities, one over relationships, one over communities

        async for message in input.message:
            query_embedding = (
                await self.embedding_provider.async_get_embedding(message)
            )

            # entity search
            search_type = "__Entity__"
            async for search_result in self.kg_provider.vector_query(  # type: ignore
                input,
                search_type=search_type,
                search_type_limits=kg_search_settings.local_search_limits[
                    search_type
                ],
                query_embedding=query_embedding,
                property_names=[
                    "name",
                    "description",
                    "fragment_ids",
                    "document_ids",
                ],
            ):
                yield KGSearchResult(
                    content=KGEntityResult(
                        name=search_result["name"],
                        description=search_result["description"],
                    ),
                    method=KGSearchMethod.LOCAL,
                    result_type=KGSearchResultType.ENTITY,
                    fragment_ids=search_result["fragment_ids"],
                    document_ids=search_result["document_ids"],
                    metadata={"associated_query": message},
                )

            # relationship search
            search_type = "__Relationship__"
            async for search_result in self.kg_provider.vector_query(  # type: ignore
                input,
                search_type=search_type,
                search_type_limits=kg_search_settings.local_search_limits[
                    search_type
                ],
                query_embedding=query_embedding,
                property_names=[
                    "name",
                    "description",
                    "fragment_ids",
                    "document_ids",
                ],
            ):
                yield KGSearchResult(
                    content=KGRelationshipResult(
                        name=search_result["name"],
                        description=search_result["description"],
                    ),
                    method=KGSearchMethod.LOCAL,
                    result_type=KGSearchResultType.RELATIONSHIP,
                    fragment_ids=search_result["fragment_ids"],
                    document_ids=search_result["document_ids"],
                    metadata={"associated_query": message},
                )

            # community search
            search_type = "__Community__"
            async for search_result in self.kg_provider.vector_query(  # type: ignore
                input,
                search_type=search_type,
                search_type_limits=kg_search_settings.local_search_limits[
                    search_type
                ],
                embedding_type="summary_embedding",
                query_embedding=query_embedding,
                property_names=["title", "summary"],
            ):

                summary = search_result["summary"]

                # try loading it as a json
                try:

                    if "```json" in summary:
                        summary = (
                            summary.strip()
                            .removeprefix("```json")
                            .removesuffix("```")
                            .strip()
                        )

                    summary_json = json.loads(summary)
                    description = summary_json.get("summary", "")
                    name = summary_json.get("title", "")

                    def get_str(finding):
                        if isinstance(finding, dict):
                            return f"{finding['summary']} => {finding['explanation']}"
                        else:
                            return str(finding)

                except json.JSONDecodeError:
                    logger.warning(f"Summary is not valid JSON")
                    continue

                yield KGSearchResult(
                    content=KGCommunityResult(
                        name=name, description=description
                    ),
                    method=KGSearchMethod.LOCAL,
                    result_type=KGSearchResultType.COMMUNITY,
                    metadata={
                        "associated_query": message,
                        "findings": summary_json.get("findings", ""),
                    },
                )

    async def global_search(
        self,
        input: GeneratorPipe.Input,
        state: AsyncState,
        run_id: UUID,
        kg_search_settings: KGSearchSettings,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[KGSearchResult, None]:
        # map reduce
        async for message in input.message:
            map_responses = []
            communities = self.kg_provider.get_communities(  # type: ignore
                level=kg_search_settings.kg_search_level
            )

            if len(communities) == 0:
                raise R2RException(
                    "No communities found. Please make sure you have run the KG enrichment step before running the search: r2r create-graph and r2r enrich-graph",
                    400,
                )

            async def preprocess_communities(communities):
                merged_report = ""
                for community in communities:
                    community_report = community.summary
                    if (
                        len(merged_report) + len(community_report)
                        > kg_search_settings.max_community_description_length
                    ):
                        yield merged_report.strip()
                        merged_report = ""
                    merged_report += community_report + "\n\n"
                if merged_report:
                    yield merged_report.strip()

            async def process_community(merged_report):
                output = await self.llm_provider.aget_completion(
                    messages=self.prompt_provider._get_message_payload(
                        task_prompt_name=self.kg_provider.config.kg_search_settings.graphrag_map_system_prompt,
                        task_inputs={
                            "context_data": merged_report,
                            "input": message,
                        },
                    ),
                    generation_config=kg_search_settings.generation_config,
                )

                return output.choices[0].message.content

            preprocessed_reports = [
                merged_report
                async for merged_report in preprocess_communities(communities)
            ]

            # Use asyncio.gather to process all preprocessed community reports concurrently
            logger.info(
                f"Processing {len(communities)} communities, {len(preprocessed_reports)} reports, Max LLM queries = {kg_search_settings.max_llm_queries_for_global_search}"
            )

            map_responses = await asyncio.gather(
                *[
                    process_community(report)
                    for report in preprocessed_reports[
                        : kg_search_settings.max_llm_queries_for_global_search
                    ]
                ]
            )
            # Filter only the relevant responses
            filtered_responses = self.filter_responses(map_responses)

            # reducing the outputs
            output = await self.llm_provider.aget_completion(
                messages=self.prompt_provider._get_message_payload(
                    task_prompt_name=self.kg_provider.config.kg_search_settings.graphrag_reduce_system_prompt,
                    task_inputs={
                        "response_type": "multiple paragraphs",
                        "report_data": filtered_responses,
                        "input": message,
                    },
                ),
                generation_config=kg_search_settings.generation_config,
            )

            output_text = output.choices[0].message.content

            if not output_text:
                logger.warning(f"No output generated for query: {message}.")
                raise R2RException(
                    "No output generated for query.",
                    400,
                )

            yield KGSearchResult(
                content=KGGlobalResult(
                    name="Global Result", description=output_text
                ),
                method=KGSearchMethod.GLOBAL,
                metadata={"associated_query": message},
            )

    async def _run_logic(  # type: ignore
        self,
        input: GeneratorPipe.Input,
        state: AsyncState,
        run_id: UUID,
        kg_search_settings: KGSearchSettings,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[KGSearchResult, None]:

        kg_search_type = kg_search_settings.kg_search_type

        # runs local and/or global search
        if kg_search_type == "local" or kg_search_type == "local_and_global":
            logger.info("Performing KG local search")
            async for result in self.local_search(
                input, state, run_id, kg_search_settings
            ):
                yield result

        if kg_search_type == "global" or kg_search_type == "local_and_global":
            logger.info("Performing KG global search")
            async for result in self.global_search(
                input, state, run_id, kg_search_settings
            ):
                yield result



import logging
import os
from contextlib import asynccontextmanager
from typing import Optional

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from .assembly import R2RBuilder, R2RConfig

logger = logging.getLogger(__name__)

# Global scheduler
scheduler = AsyncIOScheduler()


@asynccontextmanager
async def lifespan(app: FastAPI):
    # Startup
    r2r_app = await create_r2r_app(
        config_name=config_name,
        config_path=config_path,
    )

    # Copy all routes from r2r_app to app
    app.router.routes = r2r_app.app.routes

    # Copy middleware and exception handlers
    app.middleware = r2r_app.app.middleware  # type: ignore
    app.exception_handlers = r2r_app.app.exception_handlers

    # Start the scheduler
    scheduler.start()

    # Start the Hatchet worker
    await r2r_app.orchestration_provider.start_worker()

    yield

    # Shutdown
    scheduler.shutdown()


async def create_r2r_app(
    config_name: Optional[str] = "default",
    config_path: Optional[str] = None,
):
    config = R2RConfig.load(config_name, config_path)

    if (
        config.embedding.provider == "openai"
        and "OPENAI_API_KEY" not in os.environ
    ):
        raise ValueError(
            "Must set OPENAI_API_KEY in order to initialize OpenAIEmbeddingProvider."
        )

    # Build the R2RApp
    builder = R2RBuilder(config=config)
    return await builder.build()


logging.basicConfig(level=logging.INFO)

config_name = os.getenv("CONFIG_NAME", None)
config_path = os.getenv("CONFIG_PATH", None)
if not config_path and not config_name:
    config_name = "default"
host = os.getenv("HOST", "0.0.0.0")
port = int(os.getenv("PORT", "7272"))

logger.info(
    f"Environment CONFIG_NAME: {'None' if config_name is None else config_name}"
)
logger.info(
    f"Environment CONFIG_PATH: {'None' if config_path is None else config_path}"
)

# Create the FastAPI app
app = FastAPI(lifespan=lifespan)

# Add CORS middleware
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)



# type: ignore
from .csv_parser import CSVParser, CSVParserAdvanced
from .json_parser import JSONParser
from .xlsx_parser import XLSXParser, XLSXParserAdvanced

__all__ = [
    "CSVParser",
    "CSVParserAdvanced",
    "JSONParser",
    "XLSXParser",
    "XLSXParserAdvanced",
]
